#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Options for the config validator agent"""
from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging

GROUP = cfg.OptGroup(
    name='agent',
    title='Congress agent options for config datasource')

AGT_OPTS = [
    cfg.StrOpt('host', required=True),
    cfg.StrOpt('version', required=True, help='OpenStack version'),
    cfg.IntOpt('max_delay', default=10, help='The maximum delay an agent will '
                                             'wait before sending his files. '
                                             'The smaller the value, the more '
                                             'likely congestion is to happen'
                                             '.'),

    cfg.Opt(
        'services',
        help='Services activated on this node and configuration files',
        default={},
        sample_default=(
            'nova: { /etc/nova/nova.conf:/path1.conf }, '
            'neutron: { /etc/nova/neutron.conf:/path2.conf },'),
        type=types.Dict(
            bounds=False,
            value_type=types.Dict(bounds=True, value_type=types.String()))),

]


def register_validator_agent_opts(conf):
    """Register the options of the agent in the config object"""
    conf.register_group(GROUP)
    conf.register_opts(AGT_OPTS, group=GROUP)
    logging.register_options(conf)


def list_opts():
    """List agent options"""
    return [(GROUP, AGT_OPTS)]



# Copyright (c) 2018 Orange. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""A static Datalog typechecker.

It is mandatory for Z3, but should be usable by the standard engine.
Everything should be typable as we can always assign CongressStr as
a catch-all type.

The typechecker works as follows:

* Types are stored in theories schema.

* First we reset all types to bottom in theories that we want to type-check.
  The type-checker must not change the types in other theories.
  The bottom type is None with nullability set to False.
  It must be set explicitly because otherwise the default type is scalar, True
  which corresponds to the top type in the type hierarchy.

* We prepare a type environment for each rule (identified by rule.id).that
  contains the type of the variables (again a cell initialized to the bottom
  type)

* First we type facts. Then we type rules iteratively.
  A work phase types all the rules in the theories. Typing a rule means
  propagating constraints in atoms. head and body atoms are treated with
  the same algorithm.

* We request types to be the same when we solve a constraint or to be in
  direct subtype relation. In that case we take the most precise type.
  The reason for this is that we do not precisely constrain constants in
  programs. Their type will be forced by the constraints from external tables.
  Verification will occur when we translate the value to Z3.

* Types from external tables cannot be constrained. If the type of an
  external table should be changed, typing fails.

* built-ins support will be added in the future. It is the only tricky point
  as we usually need polymorphism for equalities, tests, some numeric
  operations that can operate on various size of integers.

* convergence: there is a finite number of type cells (one per column for
  tables and one per variable for rules). We iterate only if the type of
  at least one type cell has been made more precise. There are only finite
  ascending chains of types if the type hierarchy is well founded (and it is
  ultimately based on python inheritance hierarchy which is well-founded).
"""

from oslo_log import log as logging
import six

from congress import data_types
from congress.datalog import compile as ast
from congress import exception

MYPY = False
# pylint: disable = ungrouped-imports
if MYPY:
    # pylint: disable = unused-import
    from mypy_extensions import TypedDict  # noqa
    from typing import Any, Union, List, Dict, Optional  # noqa

    from congress.datalog import base   # noqa

    CELLTYPE = TypedDict(
        'CELLTYPE', {'nullable': bool, 'type': Optional[str]})

LOG = logging.getLogger(__name__)


def min_type(name1, name2, strict):
    # type: (str, str, boolean) -> Optional[str]
    """Given two type names, gives back the most precise one or None.

    If one of the type is more precise than the other, give it back otherwise
    gives back None. strict implies that the second type cannot be
    constrained. Usually because it is defined in an external table.
    """
    typ1 = data_types.TypesRegistry.type_class(name1)
    typ2 = data_types.TypesRegistry.type_class(name2)
    if typ2.least_ancestor([typ1]) is not None:
        return name2
    if not strict and typ1.least_ancestor([typ2]) is not None:
        return name1
    return None


class Typechecker(object):
    """Typechecks a set of theories"""

    def __init__(self, theories, world):
        # type: (List[base.Theory], Dict[str, base.Theory]) -> None
        self.world = world
        self.theories = theories
        self.theorynames = set(th.name for th in theories)
        self.work = False
        self.once = False
        self.type_env = {}   # type: Dict[str, Dict[str, CELLTYPE]]

    def constrain_type(self, cell, typ):
        # type: (CELLTYPE, str) -> None
        """Constrains the type set in a type cell"""
        if cell['type'] is None:
            cell['type'] = typ
            self.work = True
        else:
            old_typ = cell['type']
            if typ != old_typ:
                cell['type'] = 'Scalar'
                self.work = True

    def set_nullable(self, cell):
        # type: (CELLTYPE) -> None
        """Force type to be nullable"""
        if not cell['nullable']:
            cell['nullable'] = True
            self.work = True

    def type_cells(self, cell1, cell2, strict):
        # type: (CELLTYPE, Union[str, CELLTYPE]) -> Optional[str]
        """Propagates type constraints between two type cells

        Updates work if a change has been made.

        :param cell1: type cell to constrain
        :param cell2: type cell to constrain
        :param strict: boolean, true if cell2 is from an external table and
             cannot be changed.
        :return: None if ok, the text of an error otherwise.
        """
        if isinstance(cell2, six.string_types):
            # Just fake the cells. Occurs for a table from a nonrecursive
            # theory.
            cell2 = {'type': 'Scalar', 'nullable': True}
        if (cell1['nullable'] and not cell2.get('nullable', True)
                and not strict):
            cell2['nullable'] = True
            self.work = True
        if cell2.get('nullable', True) and not cell1['nullable']:
            cell1['nullable'] = True
            self.work = True
        typ1 = cell1['type']
        typ2 = cell2['type']
        if typ1 is None and typ2 is not None:
            cell1['type'] = typ2
            self.work = True
        elif typ1 is not None and typ2 is None:
            cell2['type'] = typ1
            self.work = True
        elif typ1 is not None:  # then typ2 is not None too
            if typ1 != typ2:
                typ3 = min_type(typ1, typ2, strict)
                if typ3 is not None:
                    cell1['type'] = typ3
                    cell2['type'] = typ3
                    self.work = True
                else:
                    return "{} != {}".format(typ1, typ2)
        # else: two unresolved constraints, we do nothing
        return None

    def type_constant(self, value, column):
        # type: (Any, CELLTYPE) -> None
        """Types a constant and set the constraint"""
        if value is None:
            self.set_nullable(column)
        elif isinstance(value, six.string_types):
            self.constrain_type(column, 'Str')
        elif isinstance(value, bool):
            self.constrain_type(column, 'Bool')
        elif isinstance(value, int):
            self.constrain_type(column, 'Int')
        elif isinstance(value, float):
            self.constrain_type(column, 'Float')
        else:
            self.constrain_type(column, 'Scalar')

    def reset_type_environment(self):
        """Reset the type environment for all variables in rules"""
        self.type_env = {
            rule.id: {
                variable.name: {'type': None, 'nullable': False}
                for variable in rule.variables()
            }
            for theory in self.theories
            for ruleset in theory.rules.rules.values()
            for rule in ruleset
        }

    def reset_types(self):
        """Set all types in theory to typechecks to bottom"""

        def refresh_item(elt):
            """Refresh the type of a table's column"""
            elt = {'name': elt} if isinstance(elt, six.string_types) else elt
            elt['type'] = None
            elt['nullable'] = False
            return elt
        for theory in self.theories:
            theory.schema.map = {
                k: [refresh_item(e) for e in row]
                for (k, row) in six.iteritems(theory.schema.map)}

    def type_facts(self, theory):
        # type: (base.Theory) -> None
        """Types the facts taking the best plausible type from arguments"""
        for (tablename, facts) in six.iteritems(theory.rules.facts):
            type_row = theory.schema.map[tablename]
            for fact in facts:
                for (value, typ) in six.moves.zip(fact, type_row):
                    self.type_constant(value, typ)

    def type_rule(self, theory, rule):
        # type: (base.Theory, ast.Rule) -> None
        """One type iteration over a single rule"""
        LOG.debug("Type rule %s", rule.id)
        var_types = self.type_env[rule.id]

        def type_atom(atom):
            # type: (ast.Literal) -> None
            """Type iteration for a single atom"""
            table = atom.table
            svc = theory.name if table.service is None else table.service
            tablename = table.table
            if svc == 'builtin':
                raise exception.PolicyRuntimeException(
                    'typing Z3 theories with builtin not supported yet')
            strict = svc not in self.theorynames
            tbl_schema = self.world[svc].schema.map[tablename]
            for (arg, typ_col) in six.moves.zip(atom.arguments, tbl_schema):
                if isinstance(arg, ast.Variable):
                    typ_var = var_types[arg.name]
                    err = self.type_cells(typ_var, typ_col, strict)
                    if err is not None:
                        raise exception.PolicyRuntimeException(
                            ("Type error while typing variable '{}' "
                             "in {} in rule {}: {}").format(
                                 arg.name, atom, rule.id, err)
                        )
                elif isinstance(arg, ast.ObjectConstant) and self.once:
                    self.type_constant(arg.name, typ_col)

        for atom in rule.body:
            type_atom(atom)
        type_atom(rule.head)

    def type_all(self):
        """Iterative typechecker"""
        self.reset_types()
        self.reset_type_environment()
        for theory in self.theories:
            self.type_facts(theory)

        self.work = True
        self.once = True
        while self.work:
            LOG.debug("*** Z3 Type iteration")
            self.work = False
            for theory in self.theories:
                for ruleset in theory.rules.rules.values():
                    for rule in ruleset:
                        self.type_rule(theory, rule)
            self.once = False



# Copyright (c) 2017 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.api import base
from congress.api import error_codes
from congress.api import webservice
from congress import exception

LOG = logging.getLogger(__name__)


class LibraryPolicyModel(base.APIModel):
    """Model for handling API requests about Library Policies."""

    # Note(thread-safety): blocking function
    def get_items(self, params, context=None):
        """Get items in model.

        :param: params: A dict-like object containing parameters
                    from the request query string and body.
                    The name parameter filters results by name policy name.
        :param: context: Key-values providing frame of reference of request

        :returns: A dict containing at least a 'results' key whose value is
                 a list of items in the model.  Additional keys set in the
                 dict will also be rendered for the user.
        """
        include_rules = True
        if params.get('include_rules', 'true').lower() == 'false':
            include_rules = False
        try:
            # Note: name is included as a filtering parameter in get_items
            # rather than a key in get_item because the API does not commit to
            # library policy name being unique.
            if 'name' in params:
                # Note(thread-safety): blocking call
                try:
                    policy = self.invoke_rpc(
                        base.LIBRARY_SERVICE_ID, 'get_policy_by_name',
                        {'name': params['name'],
                         'include_rules': include_rules})
                    return {"results": [policy]}
                except KeyError:  # not found
                    return {"results": []}
            else:
                # Note(thread-safety): blocking call
                return {"results": self.invoke_rpc(
                    base.LIBRARY_SERVICE_ID,
                    'get_policies', {'include_rules': include_rules})}
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)

    # Note(thread-safety): blocking function
    def get_item(self, id_, params, context=None):
        """Retrieve item with id from model.

        :param: id\_: The id of the item to retrieve
        :param: params: A dict-like object containing parameters
                    from the request query string and body.
        :param: context: Key-values providing frame of reference of request

        :returns: The matching item or None if no item with id exists.
        """
        try:
            # Note(thread-safety): blocking call
            include_rules = True
            if params.get('include_rules', 'true').lower() == 'false':
                include_rules = False
            return self.invoke_rpc(base.LIBRARY_SERVICE_ID,
                                   'get_policy',
                                   {'id_': id_,
                                    'include_rules': include_rules})
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)

    # Note(thread-safety): blocking function
    def add_item(self, item, params, id_=None, context=None):
        """Add item to model.

        :param: item: The item to add to the model
        :param: params: A dict-like object containing parameters
                    from the request query string and body.
        :param: id\_: The unique name of the item
        :param: context: Key-values providing frame of reference of request

        :returns: Tuple of (ID, newly_created_item)

        :raises KeyError: ID already exists.
        :raises DataModelException: Addition cannot be performed.
        """
        if id_ is not None:
            (num, desc) = error_codes.get('policy_id_must_not_be_provided')
            raise webservice.DataModelException(num, desc)

        try:
            # Note(thread-safety): blocking call
            policy_metadata = self.invoke_rpc(
                base.LIBRARY_SERVICE_ID, 'create_policy',
                {'policy_dict': item})
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)

        return (policy_metadata['id'], policy_metadata)

    # Note(thread-safety): blocking function
    def delete_item(self, id_, params, context=None):
        """Remove item from model.

        :param: id\_: The unique name of the item to be removed
        :param: params:
        :param: context: Key-values providing frame of reference of request

        :returns: The removed item.

        :raises KeyError: Item with specified id\_ not present.
        """
        # Note(thread-safety): blocking call
        return self.invoke_rpc(base.LIBRARY_SERVICE_ID,
                               'delete_policy',
                               {'id_': id_})

    def replace_item(self, id_, item, params, context=None):
        """Replace item with id\_ with new data.

        :param: id\_: The ID of the item to be replaced
        :param: item: The new item
        :param: params: A dict-like object containing parameters
                    from the request query string and body.
        :param: context: Key-values providing frame of reference of request

        :returns: The new item after replacement.

        :raises KeyError: Item with specified id\_ not present.
        """
        # Note(thread-safety): blocking call
        try:
            return self.invoke_rpc(base.LIBRARY_SERVICE_ID,
                                   'replace_policy',
                                   {'id_': id_,
                                    'policy_dict': item})
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)



# Copyright (c) 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import


policy_data = """
{
    "context_is_admin":  "role:admin",
    "admin_only": "rule:context_is_admin",
    "regular_user": "",
    "default": "rule:admin_only"
}
"""



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy


rules = [
    policy.RuleDefault(
        name='context_is_admin',
        check_str='role:admin'
    ),
    policy.RuleDefault(
        name='admin_only',
        check_str='rule:context_is_admin'
    ),
    policy.RuleDefault(
        name='regular_user',
        check_str='',
        description='The policy rule defining who is a regular user. This '
                    'rule can be overridden by, for example, a role check.'
    ),
    policy.RuleDefault(
        name='default',
        check_str='rule:admin_only',
        description='The default policy rule to apply when enforcing API '
                    'permissions. By default, all APIs are admin only. '
                    'This rule can be overridden (say by rule:regular_user) '
                    'to allow non-admins to access Congress APIs.'
    )
]


def list_rules():
    return rules



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import functools
import json
import sys

from oslo_log import log as logging

from congress.api import application
from congress import harness

LOG = logging.getLogger(__name__)


def fail_gracefully(f):
    """Logs exceptions and aborts."""
    @functools.wraps(f)
    def wrapper(*args, **kw):
        try:
            return f(*args, **kw)
        except Exception:
            LOG.exception("Fatal Exception:")
            sys.exit(1)

    return wrapper


@fail_gracefully
def congress_app_factory(global_conf, **local_conf):
    # global_conf only accepts an iteratable value as its dict value
    flags_dict = json.loads(global_conf['flags'])
    services = harness.create2(
        node_id=global_conf['node_id'],
        bus_id=global_conf['bus_id'],
        policy_engine=flags_dict['policy_engine'],
        api=flags_dict['api'],
        datasources=flags_dict['datasources'])
    return application.ApiApplication(services['api_service'])



# Copyright (c) 2016 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from futurist import periodics
import mock
from oslo_config import cfg

from congress.api import base as api_base
from congress.common import config
from congress import harness
from congress.tests import fake_datasource
from congress.tests import helper


def setup_config(with_fake_datasource=True, node_id='testnode',
                 same_partition_as_node=None, api=True, policy=True,
                 datasources=True):
    """Setup DseNode for testing.

    :param: services is an array of DataServices
    :param: api is a dictionary mapping api name to API model instance
    """
    config.set_config_defaults()
    # Load the fake driver.
    cfg.CONF.set_override(
        'drivers',
        ['congress.tests.fake_datasource.FakeDataSource'])

    if same_partition_as_node is None:
        node = helper.make_dsenode_new_partition(node_id)
    else:
        node = helper.make_dsenode_same_partition(
            same_partition_as_node, node_id)

    if datasources:
        cfg.CONF.set_override('datasources', True)

    with mock.patch.object(periodics, 'PeriodicWorker', autospec=True):
        services = harness.create2(
            existing_node=node, policy_engine=policy, api=api,
            datasources=datasources)

    data = None
    if with_fake_datasource:
        data = fake_datasource.FakeDataSource('data')
        # FIXME(ekcs): this is a hack to prevent the synchronizer from
        # attempting to delete this DSD because it's not in DB
        data.type = 'no_sync_datasource_driver'
        node.register_service(data)

    engine_service = None
    library_service = None
    api_service = None
    if policy:
        engine_service = services[api_base.ENGINE_SERVICE_ID]
        library_service = services[api_base.LIBRARY_SERVICE_ID]
    if api:
        api_service = services['api']
    if datasources:
        ds_manager = services['ds_manager']

    return {'node': node, 'engine': engine_service, 'library': library_service,
            'data': data, 'api': api_service, 'ds_manager': ds_manager}



# Copyright (c) 2016 NEC Corp. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import eventlet
from futurist import periodics
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from congress.datalog import base
from congress.datalog import compile
from congress.db import datasources
from congress.db import db_policy_rules

LOG = logging.getLogger(__name__)

SYNCHRONIZER_SERVICE_ID = '_policy_rule_synchronizer'


class PolicyRuleSynchronizer(object):

    def __init__(self, service_obj, node):
        self.name = SYNCHRONIZER_SERVICE_ID
        self.engine = service_obj
        self.sync_thread = None
        self.periodic_tasks = None
        self.node = node

    def start(self):
        callables = [(self.synchronize_all_policies, None, {}),
                     (self.synchronize_rules, None, {})]
        self.periodic_tasks = periodics.PeriodicWorker(callables)
        self.sync_thread = eventlet.spawn_n(self.periodic_tasks.start)
        LOG.info("started policy-rule synchronizer on node %s",
                 self.node.node_id)

    def stop(self):
        if self.periodic_tasks:
            self.periodic_tasks.stop()
            self.periodic_tasks.wait()
            self.periodic_tasks = None
        if self.sync_thread is not None:
            eventlet.greenthread.kill(self.sync_thread)
            self.sync_thread = None

    def _register_datasource_with_pe(self, ds_name):
        """create datasource policy in PE for newly created datasource."""
        if not self.node.is_valid_service(ds_name):
            # datasource service not up, nothing to register
            return
        # Get the datasource schema to sync the schema with PE
        schema = self.node.invoke_service_rpc(ds_name, 'get_datasource_schema',
                                              {'source_id': ds_name})
        self.engine.initialize_datasource(ds_name, schema)
        LOG.debug("registered datasource '%s' with PE on node %s", ds_name,
                  self.node.node_id)

    def _sync_datasource_policies(self):
        added = 0
        removed = 0
        db_datasources = [ds['name'] for ds in self.node.get_datasources()]
        ds_policies = [p['name'] for p in
                       self._get_engine_policies(datasource=True)]

        for ds in db_datasources:
            # check if ds is registered with PE
            if ds not in ds_policies:
                self._register_datasource_with_pe(ds)
                added = added + 1

        # get the policies registered with PE , but not in database
        remove_policies = list(set(ds_policies) - set(db_datasources))
        for p in remove_policies:
            self.engine.delete_policy(p)
            removed = removed+1

        LOG.debug("datasource policies synchronized, added %d removed %d ",
                  added, removed)

    def _get_engine_policies(self, datasource=False):
        all_policies = [self.engine.policy_object(n) for n in
                        self.engine.policy_names()]
        dpolicies = [p for p in all_policies
                     if p.kind == base.DATASOURCE_POLICY_TYPE]
        epolicies = list(set(all_policies) - set(dpolicies))
        policies = dpolicies if datasource else epolicies
        active_policies = []
        for policy in policies:
            active_policies.append({'id': policy.id,
                                    'name': policy.name,
                                    'abbreviation': policy.abbr,
                                    'description': policy.desc,
                                    'owner_id': policy.owner,
                                    'kind': policy.kind})

        return active_policies

    @lockutils.synchronized('congress_synchronize_policies')
    def sync_one_policy(self, name, datasource=True, db_session=None):
        return self.sync_one_policy_nonlocking(
            name, datasource=datasource, db_session=db_session)

    def sync_one_policy_nonlocking(
            self, name, datasource=True, db_session=None):
        """Synchronize single policy with DB.

        :param: name: policy name to be synchronized
        :param: datasource: True, to sync a datasource policy

        """
        LOG.info("sync %s policy with DB", name)

        if datasource:
            policy_object = datasources.get_datasource_by_name(
                name, session=db_session)
            if policy_object is not None:
                if name not in self.engine.policy_names():
                    self._register_datasource_with_pe(name)
                return

        policy_object = db_policy_rules.get_policy_by_name(
            name, session=db_session)
        if policy_object is None:
            if name in self.engine.policy_names():
                self.engine.delete_policy(name)
                LOG.info("policy %s deleted by synchronizer", name)
            return

        p = policy_object.to_dict()
        if name not in self.engine.policy_names():
            self.engine.create_policy(
                p['name'], id_=p['id'], abbr=p['abbreviation'],
                kind=p['kind'], desc=p['description'],
                owner=p['owner_id'])
            LOG.debug("policy %s added by synchronizer", name)

        elif p['id'] != self.engine.policy_object(name).id:
            # if same name but not identical attributes
            # replace by new policy obj according to DB
            self.engine.delete_policy(name)
            self.engine.create_policy(
                p['name'], id_=p['id'], abbr=p['abbreviation'],
                kind=p['kind'], desc=p['description'],
                owner=p['owner_id'])
            LOG.debug("synchronizer, policy replaced %s", name)

    @periodics.periodic(spacing=cfg.CONF.datasource_sync_period)
    @lockutils.synchronized('congress_synchronize_policies')
    def synchronize_all_policies(self):
        """Function to synchronize im-mem policies with DB"""
        added = 0
        removed = 0
        try:
            db_policies = [p.to_dict() for p in db_policy_rules.get_policies()]
            active_policies = self._get_engine_policies()
            # Delete engine policies which are not in DB
            for p in active_policies:
                if p not in db_policies:
                    LOG.debug("removing policy %s", str(p))
                    self.engine.delete_policy(p['id'])
                    removed = removed + 1
            # Add policies to PE, which are in DB
            for p in db_policies:
                if p not in active_policies:
                    LOG.debug("adding policy %s", str(p))
                    self.engine.create_policy(p['name'], id_=p['id'],
                                              abbr=p['abbreviation'],
                                              kind=p['kind'],
                                              desc=p['description'],
                                              owner=p['owner_id'])
                    added = added + 1
            LOG.info("engine policies synchronized, added %d removed %d ",
                     added, removed)
            # synchronize datasource policies
            self._sync_datasource_policies()
            LOG.info("completed synchronization of policies")
        except Exception:
            LOG.exception("Exception occurred in policy synchronizer periodic"
                          "task on node %s", self.node.node_id)
            return

    @periodics.periodic(spacing=cfg.CONF.datasource_sync_period)
    @lockutils.synchronized('congress_synchronize_rules')
    def synchronize_rules(self, db_session=None):
        self.synchronize_rules_nonlocking(db_session=db_session)

    def synchronize_rules_nonlocking(self, db_session=None):
        LOG.debug("Synchronizing rules on node %s", self.node.node_id)
        try:
            # Read rules from DB.
            configured_rules = []
            configured_facts = []
            for r in db_policy_rules.get_policy_rules(session=db_session):
                if ':-' in r.rule:  # if rule has body
                    configured_rules.append({'rule': r.rule,
                                             'id': r.id,
                                             'comment': r.comment,
                                             'name': r.name,
                                             'policy_name': r.policy_name})
                else:  # head-only rule, ie., fact
                    configured_facts.append(
                        {'rule': self.engine.parse1(r.rule).pretty_str(),
                         # note:parse to remove effect of extraneous formatting
                         'policy_name': r.policy_name})

            # Read rules from engine
            policies = {n: self.engine.policy_object(n) for n in
                        self.engine.policy_names()}
            active_policy_rules = []
            active_policy_facts = []
            for policy_name, policy in policies.items():
                if policy.kind != base.DATASOURCE_POLICY_TYPE:
                    for active_rule in policy.content():
                        # FIXME: This assumes r.original_str is None iff
                        # r is a head-only rule (fact). This works in
                        # non-recursive policy but not in recursive policies
                        if active_rule.original_str is None:
                            active_policy_facts.append(
                                {'rule': str(active_rule.head),
                                 'policy_name': policy_name})
                        else:
                            active_policy_rules.append(
                                {'rule': active_rule.original_str,
                                 'id': active_rule.id,
                                 'comment': active_rule.comment,
                                 'name': active_rule.name,
                                 'policy_name': policy_name})

            # ALEX: the Rule object does not have fields like the rule-string
            # or id or comment.  We can add those fields to the Rule object,
            # as long as we don't add them to the Fact because there are many
            # fact instances.  If a user tries to create a lot of Rules, they
            # are probably doing something wrong and should use a datasource
            # driver instead.

            changes = []

            # add configured rules
            for r in configured_rules:
                if r not in active_policy_rules:
                    LOG.debug("adding rule %s", str(r))
                    parsed_rule = self.engine.parse1(r['rule'])
                    parsed_rule.set_id(r['id'])
                    parsed_rule.set_name(r['name'])
                    parsed_rule.set_comment(r['comment'])
                    parsed_rule.set_original_str(r['rule'])

                    event = compile.Event(formula=parsed_rule,
                                          insert=True,
                                          target=r['policy_name'])
                    changes.append(event)

            # add configured facts
            for r in configured_facts:
                if r not in active_policy_facts:
                    LOG.debug("adding rule %s", str(r))
                    parsed_rule = self.engine.parse1(r['rule'])
                    event = compile.Event(formula=parsed_rule,
                                          insert=True,
                                          target=r['policy_name'])
                    changes.append(event)

            # remove active rules not configured
            for r in active_policy_rules:
                if r not in configured_rules:
                    LOG.debug("removing rule %s", str(r))
                    parsed_rule = self.engine.parse1(r['rule'])
                    parsed_rule.set_id(r['id'])
                    parsed_rule.set_name(r['name'])
                    parsed_rule.set_comment(r['comment'])
                    parsed_rule.set_original_str(r['rule'])

                    event = compile.Event(formula=parsed_rule,
                                          insert=False,
                                          target=r['policy_name'])
                    changes.append(event)

            # remove active facts not configured
            for r in active_policy_facts:
                if r not in configured_facts:
                    LOG.debug("removing rule %s", str(r))
                    parsed_rule = self.engine.parse1(r['rule'])
                    event = compile.Event(formula=parsed_rule,
                                          insert=False,
                                          target=r['policy_name'])
                    changes.append(event)

            permitted, changes = self.engine.process_policy_update(changes)
            LOG.info("synchronize_rules, permitted %d, made %d changes on "
                     "node %s", permitted, len(changes), self.node.node_id)
        except Exception:
            LOG.exception("synchronizing rules failed")



# Copyright (c) 2018 NEC, Corp
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils


class TestCustomDriver(datasource_driver.PollingDataSourceDriver):

    def __init__(self, name='', args=None):
        super(TestCustomDriver, self).__init__(name, args)

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'test'
        result['description'] = 'This is a fake driver used for testing'
        result['config'] = datasource_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result



#!/usr/bin/env python
# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import neutronclient.v2_0.client
from oslo_log import log as logging

from congress import data_types
from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


IngressEgress = data_types.create_congress_enum_type(
    'IngressEgress', ('ingress', 'egress'), data_types.Str)
data_types.TypesRegistry.register(IngressEgress)

FloatingIPStatus = data_types.create_congress_enum_type(
    'FloatingIPStatus', ('ACTIVE', 'DOWN', 'ERROR'), data_types.Str,
    catch_all_default_value='OTHER')
data_types.TypesRegistry.register(FloatingIPStatus)

NeutronStatus = data_types.create_congress_enum_type(
    'NeutronStatus', ('ACTIVE', 'DOWN', 'BUILD', 'ERROR'), data_types.Str,
    catch_all_default_value='OTHER')
data_types.TypesRegistry.register(NeutronStatus)

IPVersion = data_types.create_congress_enum_type(
    'IPv4IPv6', (4, 6), data_types.Int)
data_types.TypesRegistry.register(IPVersion)


class NeutronV2Driver(datasource_driver.PollingDataSourceDriver,
                      datasource_driver.ExecutionDriver):

    NETWORKS = 'networks'
    FIXED_IPS = 'fixed_ips'
    SECURITY_GROUP_PORT_BINDINGS = 'security_group_port_bindings'
    PORTS = 'ports'
    ALLOCATION_POOLS = 'allocation_pools'
    DNS_NAMESERVERS = 'dns_nameservers'
    HOST_ROUTES = 'host_routes'
    SUBNETS = 'subnets'
    EXTERNAL_FIXED_IPS = 'external_fixed_ips'
    EXTERNAL_GATEWAY_INFOS = 'external_gateway_infos'
    ROUTERS = 'routers'
    SECURITY_GROUP_RULES = 'security_group_rules'
    SECURITY_GROUPS = 'security_groups'
    FLOATING_IPS = 'floating_ips'

    value_trans_str = ds_utils.typed_value_trans(data_types.Str)
    value_trans_bool = ds_utils.typed_value_trans(data_types.Bool)
    value_trans_int = ds_utils.typed_value_trans(data_types.Int)

    floating_ips_translator = {
        'translation-type': 'HDICT',
        'table-name': FLOATING_IPS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID of the floating IP address',
              'translator': value_trans_str},
             {'fieldname': 'router_id', 'desc': 'UUID of router',
              'translator': value_trans_str},
             {'fieldname': 'tenant_id', 'desc': 'Tenant ID',
              'translator': value_trans_str},
             {'fieldname': 'floating_network_id',
              'desc': 'The UUID of the network associated with floating IP',
              'translator': value_trans_str},
             {'fieldname': 'fixed_ip_address',
              'desc': 'Fixed IP address associated with floating IP address',
              'translator': ds_utils.typed_value_trans(data_types.IPAddress)},
             {'fieldname': 'floating_ip_address',
              'desc': 'The floating IP address',
              'translator': ds_utils.typed_value_trans(data_types.IPAddress)},
             {'fieldname': 'port_id', 'desc': 'UUID of port',
              'translator': value_trans_str},
             {'fieldname': 'status', 'desc': 'The floating IP status',
              'translator': ds_utils.typed_value_trans(FloatingIPStatus)})}

    networks_translator = {
        'translation-type': 'HDICT',
        'table-name': NETWORKS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'Network ID',
              'translator': value_trans_str},
             {'fieldname': 'tenant_id', 'desc': 'Tenant ID',
              'translator': value_trans_str},
             {'fieldname': 'name', 'desc': 'Network name',
              'translator': value_trans_str},
             {'fieldname': 'status', 'desc': 'Network status',
              'translator': ds_utils.typed_value_trans(NeutronStatus)},
             {'fieldname': 'admin_state_up',
              'desc': 'Administrative state of the network (true/false)',
              'translator': value_trans_bool},
             {'fieldname': 'shared',
              'desc': 'Indicates if network is shared across all tenants',
              'translator': value_trans_bool})}

    ports_fixed_ips_translator = {
        'translation-type': 'HDICT',
        'table-name': FIXED_IPS,
        'parent-key': 'id',
        'parent-col-name': 'port_id',
        'parent-key-desc': 'UUID of Port',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'ip_address',
              'desc': 'The IP addresses for the port',
              'translator': ds_utils.typed_value_trans(data_types.IPAddress)},
             {'fieldname': 'subnet_id',
              'desc': 'The UUID of the subnet to which the port is attached',
              'translator': value_trans_str})}

    ports_security_groups_translator = {
        'translation-type': 'LIST',
        'table-name': SECURITY_GROUP_PORT_BINDINGS,
        'parent-key': 'id',
        'parent-col-name': 'port_id',
        'parent-key-desc': 'UUID of port',
        'val-col': 'security_group_id',
        'val-col-desc': 'UUID of security group',
        'translator': value_trans_str}

    ports_translator = {
        'translation-type': 'HDICT',
        'table-name': PORTS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'UUID of port',
              'translator': value_trans_str},
             {'fieldname': 'tenant_id', 'desc': 'tenant ID',
              'translator': value_trans_str},
             {'fieldname': 'name', 'desc': 'port name',
              'translator': value_trans_str},
             {'fieldname': 'network_id', 'desc': 'UUID of attached network',
              'translator': value_trans_str},
             {'fieldname': 'mac_address', 'desc': 'MAC address of the port',
              'translator': value_trans_str},
             {'fieldname': 'admin_state_up',
              'desc': 'Administrative state of the port',
              'translator': value_trans_bool},
             {'fieldname': 'status', 'desc': 'Port status',
              'translator': ds_utils.typed_value_trans(NeutronStatus)},
             {'fieldname': 'device_id',
              'desc': 'The ID of the device that uses this port',
              'translator': value_trans_str},
             {'fieldname': 'device_owner',
              'desc': 'The entity type that uses this port.'
                      'E.g., compute:nova, network:router_interface',
              'translator': value_trans_str},
             {'fieldname': 'fixed_ips',
              'desc': 'The IP addresses for the port',
              'translator': ports_fixed_ips_translator},
             {'fieldname': 'security_groups',
              'translator': ports_security_groups_translator})}

    subnets_allocation_pools_translator = {
        'translation-type': 'HDICT',
        'table-name': ALLOCATION_POOLS,
        'parent-key': 'id',
        'parent-col-name': 'subnet_id',
        'parent-key-desc': 'UUID of subnet',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'start',
              'desc': 'The start address for the allocation pools',
              'translator': value_trans_str},
             {'fieldname': 'end',
              'desc': 'The end address for the allocation pools',
              'translator': value_trans_str})}

    subnets_dns_nameservers_translator = {
        'translation-type': 'LIST',
        'table-name': DNS_NAMESERVERS,
        'parent-key': 'id',
        'parent-col-name': 'subnet_id',
        'parent-key-desc': 'UUID of subnet',
        'val-col': 'dns_nameserver',
        'val-col-desc': 'The DNS server',
        'translator': value_trans_str}

    subnets_routes_translator = {
        'translation-type': 'HDICT',
        'table-name': HOST_ROUTES,
        'parent-key': 'id',
        'parent-col-name': 'subnet_id',
        'parent-key-desc': 'UUID of subnet',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'destination',
              'desc': 'The destination for static route',
              'translator': value_trans_str},
             {'fieldname': 'nexthop',
              'desc': 'The next hop for the destination',
              'translator': value_trans_str})}

    subnets_translator = {
        'translation-type': 'HDICT',
        'table-name': SUBNETS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'UUID of subnet',
              'translator': value_trans_str},
             {'fieldname': 'tenant_id', 'desc': 'tenant ID',
              'translator': value_trans_str},
             {'fieldname': 'name', 'desc': 'subnet name',
              'translator': value_trans_str},
             {'fieldname': 'network_id', 'desc': 'UUID of attached network',
              'translator': value_trans_str},
             {'fieldname': 'ip_version',
              'desc': 'The IP version, which is 4 or 6',
              'translator': ds_utils.typed_value_trans(IPVersion)},
             {'fieldname': 'cidr', 'desc': 'The CIDR',
              'translator': ds_utils.typed_value_trans(data_types.IPNetwork)},
             {'fieldname': 'gateway_ip', 'desc': 'The gateway IP address',
              'translator': ds_utils.typed_value_trans(data_types.IPAddress)},
             {'fieldname': 'enable_dhcp', 'desc': 'Is DHCP is enabled or not',
              'translator': value_trans_bool},
             {'fieldname': 'ipv6_ra_mode', 'desc': 'The IPv6 RA mode',
              'translator': value_trans_str},
             {'fieldname': 'ipv6_address_mode',
              'desc': 'The IPv6 address mode', 'translator': value_trans_str},
             {'fieldname': 'allocation_pools',
              'translator': subnets_allocation_pools_translator},
             {'fieldname': 'dns_nameservers',
              'translator': subnets_dns_nameservers_translator},
             {'fieldname': 'host_routes',
              'translator': subnets_routes_translator})}

    external_fixed_ips_translator = {
        'translation-type': 'HDICT',
        'table-name': EXTERNAL_FIXED_IPS,
        'parent-key': 'router_id',
        'parent-col-name': 'router_id',
        'parent-key-desc': 'UUID of router',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'subnet_id', 'desc': 'UUID of the subnet',
              'translator': value_trans_str},
             {'fieldname': 'ip_address', 'desc': 'IP Address',
              'translator': ds_utils.typed_value_trans(data_types.IPAddress)})}

    routers_external_gateway_infos_translator = {
        'translation-type': 'HDICT',
        'table-name': EXTERNAL_GATEWAY_INFOS,
        'parent-key': 'id',
        'parent-col-name': 'router_id',
        'parent-key-desc': 'UUID of router',
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'network_id', 'desc': 'Network ID',
              'translator': value_trans_str},
             {'fieldname': 'enable_snat',
              'desc': 'current Source NAT status for router',
              'translator': value_trans_bool},
             {'fieldname': 'external_fixed_ips',
              'translator': external_fixed_ips_translator})}

    routers_translator = {
        'translation-type': 'HDICT',
        'table-name': ROUTERS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'uuid of the router',
              'translator': value_trans_str},
             {'fieldname': 'tenant_id', 'desc': 'tenant ID',
              'translator': value_trans_str},
             {'fieldname': 'status', 'desc': 'router status',
              'translator': ds_utils.typed_value_trans(NeutronStatus)},
             {'fieldname': 'admin_state_up',
              'desc': 'administrative state of router',
              'translator': value_trans_bool},
             {'fieldname': 'name', 'desc': 'router name',
              'translator': value_trans_str},
             {'fieldname': 'distributed',
              'desc': "indicates if it's distributed router ",
              'translator': value_trans_bool},
             {'fieldname': 'external_gateway_info',
              'translator': routers_external_gateway_infos_translator})}

    security_group_rules_translator = {
        'translation-type': 'HDICT',
        'table-name': SECURITY_GROUP_RULES,
        'parent-key': 'id',
        'parent-col-name': 'security_group_id',
        'parent-key-desc': 'uuid of security group',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID of the security group rule',
              'translator': value_trans_str},
             {'fieldname': 'tenant_id', 'desc': 'tenant ID',
              'translator': value_trans_str},
             {'fieldname': 'remote_group_id',
              'desc': 'remote group id to associate with security group rule',
              'translator': value_trans_str},
             {'fieldname': 'direction',
              'desc': 'Direction in which the security group rule is applied',
              'translator': ds_utils.typed_value_trans(IngressEgress)},
             {'fieldname': 'ethertype', 'desc': 'IPv4 or IPv6',
              'translator': value_trans_str},
             {'fieldname': 'protocol',
              'desc': 'protocol that is matched by the security group rule.',
              'translator': value_trans_str},
             {'fieldname': 'port_range_min',
              'desc': 'Min port number in the range',
              'translator': value_trans_int},
             {'fieldname': 'port_range_max',
              'desc': 'Max port number in the range',
              'translator': value_trans_int},
             {'fieldname': 'remote_ip_prefix',
              'desc': 'Remote IP prefix to be associated',
              'translator': value_trans_str})}

    security_group_translator = {
        'translation-type': 'HDICT',
        'table-name': SECURITY_GROUPS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID for the security group',
              'translator': value_trans_str},
             {'fieldname': 'tenant_id', 'desc': 'Tenant ID',
              'translator': value_trans_str},
             {'fieldname': 'name', 'desc': 'The security group name',
              'translator': value_trans_str},
             {'fieldname': 'description', 'desc': 'security group description',
              'translator': value_trans_str},
             {'fieldname': 'security_group_rules',
              'translator': security_group_rules_translator})}

    TRANSLATORS = [networks_translator, ports_translator, subnets_translator,
                   routers_translator, security_group_translator,
                   floating_ips_translator]

    def __init__(self, name='', args=None):
        super(NeutronV2Driver, self).__init__(name, args=args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        session = ds_utils.get_keystone_session(self.creds)
        self.neutron = neutronclient.v2_0.client.Client(session=session)

        # specify the arg name for all method structured args
        self.method_structured_args = {
            'add_bgp_speaker_to_dragent': {'named': frozenset(['body'])},
            'add_gateway_router': {'named': frozenset(['body'])},
            'add_interface_router': {'named': frozenset(['body'])},
            'add_network_to_bgp_speaker': {'named': frozenset(['body'])},
            'add_network_to_dhcp_agent': {'named': frozenset(['body'])},
            'add_peer_to_bgp_speaker': {'named': frozenset(['body'])},
            'add_router_to_l3_agent': {'named': frozenset(['body'])},
            'associate_flavor': {'named': frozenset(['body'])},
            'associate_health_monitor': {'named': frozenset(['body'])},
            'connect_network_gateway': {'named': frozenset(['body'])},
            'create_address_scope': {'named': frozenset(['body'])},
            'create_bandwidth_limit_rule': {'named': frozenset(['body'])},
            'create_bgp_peer': {'named': frozenset(['body'])},
            'create_bgp_speaker': {'named': frozenset(['body'])},
            'create_bgpvpn': {'named': frozenset(['body'])},
            'create_bgpvpn_network_assoc': {'named': frozenset(['body'])},
            'create_bgpvpn_port_assoc': {'named': frozenset(['body'])},
            'create_bgpvpn_router_assoc': {'named': frozenset(['body'])},
            'create_dscp_marking_rule': {'named': frozenset(['body'])},
            'create_endpoint_group': {'named': frozenset(['body'])},
            'create_ext': {'named': frozenset(['body'])},
            'create_firewall': {'named': frozenset(['body'])},
            'create_firewall_policy': {'named': frozenset(['body'])},
            'create_firewall_rule': {'named': frozenset(['body'])},
            'create_flavor': {'named': frozenset(['body'])},
            'create_floatingip': {'named': frozenset(['body'])},
            'create_fwaas_firewall_group': {'named': frozenset(['body'])},
            'create_fwaas_firewall_policy': {'named': frozenset(['body'])},
            'create_fwaas_firewall_rule': {'named': frozenset(['body'])},
            'create_gateway_device': {'named': frozenset(['body'])},
            'create_health_monitor': {'named': frozenset(['body'])},
            'create_ikepolicy': {'named': frozenset(['body'])},
            'create_ipsec_site_connection': {'named': frozenset(['body'])},
            'create_ipsecpolicy': {'named': frozenset(['body'])},
            'create_lbaas_healthmonitor': {'named': frozenset(['body'])},
            'create_lbaas_l7policy': {'named': frozenset(['body'])},
            'create_lbaas_l7rule': {'named': frozenset(['body'])},
            'create_lbaas_member': {'named': frozenset(['body'])},
            'create_lbaas_pool': {'named': frozenset(['body'])},
            'create_listener': {'named': frozenset(['body'])},
            'create_loadbalancer': {'named': frozenset(['body'])},
            'create_member': {'named': frozenset(['body'])},
            'create_metering_label': {'named': frozenset(['body'])},
            'create_metering_label_rule': {'named': frozenset(['body'])},
            'create_minimum_bandwidth_rule': {'named': frozenset(['body'])},
            'create_network': {'named': frozenset(['body'])},
            'create_network_gateway': {'named': frozenset(['body'])},
            'create_network_log': {'named': frozenset(['body'])},
            'create_pool': {'named': frozenset(['body'])},
            'create_port': {'named': frozenset(['body'])},
            'create_qos_policy': {'named': frozenset(['body'])},
            'create_qos_queue': {'named': frozenset(['body'])},
            'create_rbac_policy': {'named': frozenset(['body'])},
            'create_router': {'named': frozenset(['body'])},
            'create_security_group': {'named': frozenset(['body'])},
            'create_security_group_rule': {'named': frozenset(['body'])},
            'create_service_profile': {'named': frozenset(['body'])},
            'create_sfc_flow_classifier': {'named': frozenset(['body'])},
            'create_sfc_port_chain': {'named': frozenset(['body'])},
            'create_sfc_port_pair': {'named': frozenset(['body'])},
            'create_sfc_port_pair_group': {'named': frozenset(['body'])},
            'create_sfc_service_graph': {'named': frozenset(['body'])},
            'create_subnet': {'named': frozenset(['body'])},
            'create_subnetpool': {'named': frozenset(['body'])},
            'create_trunk': {'named': frozenset(['body'])},
            'create_vip': {'named': frozenset(['body'])},
            'create_vpnservice': {'named': frozenset(['body'])},
            'disconnect_network_gateway': {'named': frozenset(['body'])},
            'firewall_policy_insert_rule': {'named': frozenset(['body'])},
            'firewall_policy_remove_rule': {'named': frozenset(['body'])},
            'insert_rule_fwaas_firewall_policy': {
                'named': frozenset(['body'])},
            'remove_interface_router': {'named': frozenset(['body'])},
            'remove_network_from_bgp_speaker': {'named': frozenset(['body'])},
            'remove_peer_from_bgp_speaker': {'named': frozenset(['body'])},
            'remove_rule_fwaas_firewall_policy': {
                'named': frozenset(['body'])},
            'replace_tag': {'named': frozenset(['body'])},
            'retry_request': {'named': frozenset(['body'])},
            'show_minimum_bandwidth_rule': {'named': frozenset(['body'])},
            'trunk_add_subports': {'named': frozenset(['body'])},
            'trunk_remove_subports': {'named': frozenset(['body'])},
        }

        self.add_executable_method('update_resource_attrs',
                                   [{'name': 'resource_type',
                                     'description': 'resource type (e.g. ' +
                                                    'port, network, subnet)'},
                                    {'name': 'id',
                                     'description': 'ID of the resource'},
                                    {'name': 'attr1',
                                     'description': 'attribute name to ' +
                                     'update (e.g. admin_state_up)'},
                                    {'name': 'attr1-value',
                                     'description': 'updated attr1 value'},
                                    {'name': 'attrN',
                                     'description': 'attribute name to ' +
                                     'update'},
                                    {'name': 'attrN-value',
                                     'description': 'updated attrN value'}],
                                   "A wrapper for update_<resource_type>()")
        self.add_executable_method('attach_port_security_group',
                                   [{'name': 'port_id',
                                     'description': 'ID of target port'},
                                    {'name': 'security_group_id',
                                     'description': 'ID security group to be '
                                                    'attached'}],
                                   "Attach a security group to port (WARNING: "
                                   "may overwrite concurrent changes to "
                                   "port's security groups list.")
        self.add_executable_method('detach_port_security_group',
                                   [{'name': 'port_id',
                                     'description': 'ID of target port'},
                                    {'name': 'security_group_id',
                                     'description': 'ID security group to be '
                                                    'detached'}],
                                   "Detach a security group to port (WARNING: "
                                   "may overwrite concurrent changes to "
                                   "port's security groups list.")

        # add action methods from client, but exclude 'update_*' because those
        # are covered by the update_resource_attr method.
        exclude_methods = ['update_address_scope', 'update_agent',
                           'update_bandwidth_limit_rule', 'update_bgp_peer',
                           'update_bgp_speaker', 'update_bgpvpn',
                           'update_bgpvpn_network_assoc',
                           'update_bgpvpn_port_assoc',
                           'update_bgpvpn_router_assoc',
                           'update_dscp_marking_rule', 'update_endpoint_group',
                           'update_ext', 'update_firewall',
                           'update_firewall_policy', 'update_firewall_rule',
                           'update_flavor', 'update_floatingip',
                           'update_fwaas_firewall_group',
                           'update_fwaas_firewall_policy',
                           'update_fwaas_firewall_rule',
                           'update_gateway_device', 'update_health_monitor',
                           'update_ikepolicy', 'update_ipsec_site_connection',
                           'update_ipsecpolicy', 'update_lbaas_healthmonitor',
                           'update_lbaas_l7policy', 'update_lbaas_l7rule',
                           'update_lbaas_member', 'update_lbaas_pool',
                           'update_listener', 'update_loadbalancer',
                           'update_member', 'update_minimum_bandwidth_rule',
                           'update_network', 'update_network_gateway',
                           'update_network_log', 'update_pool', 'update_port',
                           'update_qos_policy', 'update_quota',
                           'update_rbac_policy', 'update_router',
                           'update_security_group', 'update_service_profile',
                           'update_sfc_flow_classifier',
                           'update_sfc_port_chain', 'update_sfc_port_pair',
                           'update_sfc_port_pair_group',
                           'update_sfc_service_graph', 'update_subnet',
                           'update_subnetpool', 'update_trunk', 'update_vip',
                           'update_vpnservice']
        self.add_executable_client_methods(self.neutron,
                                           'neutronclient.v2_0.client',
                                           exclude_methods)
        self.initialize_update_methods()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'neutronv2'
        result['description'] = ('Datasource driver that interfaces with '
                                 'OpenStack Networking aka Neutron.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_methods(self):
        networks_method = lambda: self._translate_networks(
            self.neutron.list_networks())
        self.add_update_method(networks_method, self.networks_translator)

        subnets_method = lambda: self._translate_subnets(
            self.neutron.list_subnets())
        self.add_update_method(subnets_method, self.subnets_translator)

        ports_method = lambda: self._translate_ports(self.neutron.list_ports())
        self.add_update_method(ports_method, self.ports_translator)

        routers_method = lambda: self._translate_routers(
            self.neutron.list_routers())
        self.add_update_method(routers_method, self.routers_translator)

        security_method = lambda: self._translate_security_groups(
            self.neutron.list_security_groups())
        self.add_update_method(security_method,
                               self.security_group_translator)

        floatingips_method = lambda: self._translate_floating_ips(
            self.neutron.list_floatingips())
        self.add_update_method(floatingips_method,
                               self.floating_ips_translator)

    @ds_utils.update_state_on_changed(FLOATING_IPS)
    def _translate_floating_ips(self, obj):
        LOG.debug("floating_ips: %s", dict(obj))

        row_data = NeutronV2Driver.convert_objs(obj['floatingips'],
                                                self.floating_ips_translator)
        return row_data

    @ds_utils.update_state_on_changed(NETWORKS)
    def _translate_networks(self, obj):
        LOG.debug("networks: %s", dict(obj))

        row_data = NeutronV2Driver.convert_objs(obj['networks'],
                                                self.networks_translator)
        return row_data

    @ds_utils.update_state_on_changed(PORTS)
    def _translate_ports(self, obj):
        LOG.debug("ports: %s", obj)
        row_data = NeutronV2Driver.convert_objs(obj['ports'],
                                                self.ports_translator)
        return row_data

    @ds_utils.update_state_on_changed(SUBNETS)
    def _translate_subnets(self, obj):
        LOG.debug("subnets: %s", obj)
        row_data = NeutronV2Driver.convert_objs(obj['subnets'],
                                                self.subnets_translator)
        return row_data

    @ds_utils.update_state_on_changed(ROUTERS)
    def _translate_routers(self, obj):
        LOG.debug("routers: %s", obj)
        row_data = NeutronV2Driver.convert_objs(obj['routers'],
                                                self.routers_translator)
        return row_data

    @ds_utils.update_state_on_changed(SECURITY_GROUPS)
    def _translate_security_groups(self, obj):
        LOG.debug("security_groups: %s", obj)
        row_data = NeutronV2Driver.convert_objs(obj['security_groups'],
                                                self.security_group_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.neutron, action, action_args)

    def update_resource_attrs(self, args):
        positional_args = args.get('positional', [])
        if not positional_args or len(positional_args) < 4:
            LOG.error('Args for update_resource_attrs() must contain resource '
                      'type, resource ID and pairs of key-value attributes to '
                      'update')
            return

        resource_type = positional_args.pop(0)
        resource_id = positional_args.pop(0)
        action = 'update_%s' % resource_type
        update_attrs = self._convert_args(positional_args)
        body = {resource_type: update_attrs}

        action_args = {'named': {resource_type: resource_id,
                                 'body': body}}
        self._execute_api(self.neutron, action, action_args)

    def attach_port_security_group(self, args):
        self._attach_detach_port_security_group(args, attach=True)

    def detach_port_security_group(self, args):
        self._attach_detach_port_security_group(args, attach=False)

    def _attach_detach_port_security_group(self, args, attach):
        positional_args = args.get('positional', [])
        if not positional_args or len(positional_args) < 2:
            LOG.error('Args for attach_port_security_group() must contain '
                      'port id and security group id')
            return

        port_id = positional_args[0]
        security_group_id = positional_args[1]

        # get existing port security groups
        port_state = self.neutron.show_port(port_id).get('port')
        if not port_state:
            return
        port_security_groups = port_state.get('security_groups', [])

        # add/remove security group
        if security_group_id in port_security_groups:
            if attach:  # no change needed
                return
            port_security_groups.remove(security_group_id)
        else:
            if not attach:  # no change needed
                return
            port_security_groups.append(security_group_id)

        # call client to make change
        # WARNING: intervening changes to security groups binding may be lost
        body = {
            "port": {
                "security_groups": port_security_groups,
            }
        }
        self.neutron.update_port(port_id, body)



# Copyright (c) 2018 VMware
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock
from oslo_db import exception as db_exc
import testtools

from congress.db import utils as db_utils


class TestUtils(testtools.TestCase):

    def _get_fail_then_succeed_func(self, failure_exception):
        fail_then_succeed = mock.Mock(
            side_effect=[failure_exception, True])

        # set name as required by functools.wrap
        fail_then_succeed.__name__ = 'fail_then_suceed'
        return fail_then_succeed

    def test_no_retry_on_unknown_db_error(self):
        fail_then_succeed = db_utils.retry_on_db_error(
            self._get_fail_then_succeed_func(db_exc.DBError))
        self.assertRaises(db_exc.DBError, fail_then_succeed)

    def test_retry_on_db_deadlock_error(self):
        fail_then_succeed = db_utils.retry_on_db_error(
            self._get_fail_then_succeed_func(db_exc.DBDeadlock))
        self.assertTrue(fail_then_succeed())



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""poliy name uniqueness

Revision ID: aabe895bbd4d
Revises: 01e78af70b91
Create Date: 2016-11-04 13:55:05.064012

"""

# revision identifiers, used by Alembic.
revision = 'aabe895bbd4d'
down_revision = '01e78af70b91'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('policiesdeleted',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('created_at', sa.DateTime(), nullable=False),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('deleted_at', sa.DateTime(), nullable=True),
                    sa.Column('deleted', sa.String(length=36),
                              server_default='', nullable=True),
                    sa.Column('name', sa.String(length=255), nullable=False),
                    sa.Column('abbreviation', sa.String(length=5),
                              nullable=False),
                    sa.Column('description', sa.Text(), nullable=False),
                    sa.Column('owner', sa.Text(), nullable=False),
                    sa.Column('kind', sa.Text(), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')

    # copy all rows from table:policies to table:policiesdeleted
    # table:policiesdeleted used as temporary space while we recreate
    # table:policies with the right column:name type to support index/unique
    # recreate table rather than ALTER to generically support most backends
    try:
        op.execute(
            "INSERT INTO policiesdeleted "
            "SELECT * FROM policies")
    except Exception:
        # if copying of rows fail (likely because a name is longer than 255
        # stop upgrade and cleanup
        op.drop_table('policiesdeleted')
        raise

    # drop and recreate table:policies
    op.drop_table('policies')
    op.create_table('policies',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('created_at', sa.DateTime(), nullable=False),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('deleted_at', sa.DateTime(), nullable=True),
                    sa.Column('deleted', sa.String(length=36),
                              server_default='', nullable=True),
                    sa.Column('name', sa.String(length=255), nullable=False),
                    sa.Column('abbreviation', sa.String(length=5),
                              nullable=False),
                    sa.Column('description', sa.Text(), nullable=False),
                    sa.Column('owner', sa.Text(), nullable=False),
                    sa.Column('kind', sa.Text(), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')

    # copy non-(soft)deleted rows back into table:policies
    op.execute("INSERT INTO policies "
               "SELECT * FROM policiesdeleted WHERE deleted = ''")

    # delete non-(soft)deleted rows from table:policiesdeleted
    op.execute("DELETE FROM policiesdeleted WHERE deleted = ''")

    op.create_unique_constraint(None, 'policies', ['name'])


def downgrade():
    # drop and recreate table:policies with right column:name type
    # using table:policiesdeleted as temporary work space

    op.execute("INSERT INTO policiesdeleted SELECT * FROM policies")
    op.drop_table('policies')
    # op.drop_constraint(None, 'policies', type_='unique')
    op.create_table('policies',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('created_at', sa.DateTime(), nullable=False),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('deleted_at', sa.DateTime(), nullable=True),
                    sa.Column('deleted', sa.String(length=36),
                              server_default='', nullable=True),
                    sa.Column('name', sa.Text(), nullable=False),
                    sa.Column('abbreviation', sa.String(length=5),
                              nullable=False),
                    sa.Column('description', sa.Text(), nullable=False),
                    sa.Column('owner', sa.Text(), nullable=False),
                    sa.Column('kind', sa.Text(), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')

    # move all rows (deleted or not) into table:policies
    op.execute("INSERT INTO policies SELECT * FROM policiesdeleted")
    op.drop_table('policiesdeleted')



# Copyright (c) 2016 NEC Corp. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import eventlet
from futurist import periodics
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from congress.db import datasources

LOG = logging.getLogger(__name__)

SYNCHRONIZER_SERVICE_ID = '_datasource_synchronizer'


class DatasourceSynchronizer(object):

    def __init__(self, node):
        self.name = SYNCHRONIZER_SERVICE_ID
        self.sync_thread = None
        self.periodic_tasks = None
        self.node = node

    def start(self):
        callables = [(self.synchronize_all_datasources, None, {}),
                     (self._check_resub_all, None, {})]
        self.periodic_tasks = periodics.PeriodicWorker(callables)
        self.sync_thread = eventlet.spawn_n(self.periodic_tasks.start)
        LOG.info("started datasource synchronizer on node %s",
                 self.node.node_id)

    def stop(self):
        if self.periodic_tasks:
            self.periodic_tasks.stop()
            self.periodic_tasks.wait()
            self.periodic_tasks = None
        if self.sync_thread:
            eventlet.greenthread.kill(self.sync_thread)
            self.sync_thread = None

    @periodics.periodic(spacing=cfg.CONF.dse.time_to_resub)
    def _check_resub_all(self):
        LOG.debug("Running periodic resub on node %s", self.node.node_id)
        for s in self.node.get_services(True):
            s.check_resub_all()

    @lockutils.synchronized('congress_synchronize_datasources')
    def sync_datasource(self, ds_name):
        if not cfg.CONF.datasources:
            LOG.info("sync not supported on non-datasource node")
            return
        datasource = datasources.get_datasource_by_name(ds_name)
        service_obj = self.node.service_object(ds_name)

        if datasource and not service_obj:
            # register service with data node
            service = self.node.create_datasource_service(datasource)
            self.node.register_service(service)
            LOG.debug("service %s registered by synchronizer", ds_name)
            return
        if service_obj and datasource is None:
            # unregister, datasource not present in DB
            self.node.unregister_service(ds_name)
            LOG.debug("service %s unregistered by synchronizer", ds_name)
            return

    @lockutils.synchronized('congress_synchronize_datasources')
    @periodics.periodic(spacing=cfg.CONF.datasource_sync_period)
    def synchronize_all_datasources(self):
        LOG.debug("synchronizing datasources on node %s", self.node.node_id)
        added = 0
        removed = 0
        datasources = self.node.get_datasources(filter_secret=False)
        db_datasources = []
        # Look for datasources in the db, but not in the services.
        for configured_ds in datasources:
            db_datasources.append(configured_ds['id'])
            active_ds = self.node.service_object(uuid_=configured_ds['id'])
            # If datasource is not enabled, unregister the service
            if not configured_ds['enabled']:
                if active_ds:
                    LOG.debug("unregistering %s service, datasource disabled "
                              "in DB.", active_ds.service_id)
                    self.node.unregister_service(active_ds.service_id)
                    removed = removed + 1
                continue
            if active_ds is None:
                # service is not up, create the service
                LOG.debug("registering %s service on node %s",
                          configured_ds['name'], self.node.node_id)
                service = self.node.create_datasource_service(configured_ds)
                self.node.register_service(service)
                added = added + 1

        # Unregister the services which are not in DB
        active_ds_services = [s for s in self.node.get_services(True)
                              if getattr(s, 'type', '') == 'datasource_driver']
        db_datasources_set = set(db_datasources)
        stale_services = [s for s in active_ds_services
                          if s.ds_id not in db_datasources_set]
        for s in stale_services:
            LOG.debug("unregistering %s service, datasource not found in DB ",
                      s.service_id)
            self.node.unregister_service(uuid_=s.ds_id)
            removed = removed + 1

        LOG.info("synchronized datasources, added %d removed %d on node %s",
                 added, removed, self.node.node_id)

        # This might be required once we support update datasource config
        # if not self._config_eq(configured_ds, active_ds):
        #    LOG.debug('configured and active disagree: %s %s',
        #              strutils.mask_password(active_ds),
        #              strutils.mask_password(configured_ds))

        #    LOG.info('Reloading datasource: %s',
        #             strutils.mask_password(configured_ds))
        #    self.delete_datasource(configured_ds['name'],
        #                           update_db=False)
        #    self.add_datasource(configured_ds, update_db=False)

    # def _config_eq(self, db_config, active_config):
    #     return (db_config['name'] == active_config.service_id and
    #             db_config['config'] == active_config.service_info['args'])



# Copyright (c) 2018 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function

from congress.datasources import constants
from congress.datasources import datasource_utils
from congress.tests import base


class TestDatasourceDriver(base.TestCase):

    def test_get_openstack_required_config(self):
        expected_required = ['auth_url', 'password', 'project_name',
                             'username']
        expected_optional = ['endpoint', 'poll_time', 'project_domain_name',
                             'region', 'tenant_name', 'user_domain_name']
        config = datasource_utils.get_openstack_required_config()
        required = []
        optional = []
        for k, v in config.items():
            if v == constants.REQUIRED:
                required.append(k)
            elif v == constants.OPTIONAL:
                optional.append(k)
        self.assertEqual(sorted(required), expected_required)
        self.assertEqual(sorted(optional), expected_optional)



# Copyright (c) 2016 . All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from oslo_db import exception as db_exc
from oslo_log import log as logging

from congress.api import base as api_base
from congress.db import datasources as datasources_db
from congress.dse2 import data_service
from congress import exception
from congress.synchronizer import datasource_synchronizer

LOG = logging.getLogger(__name__)


class DSManagerService(data_service.DataService):
    """A proxy service to datasource managing methods in dse_node."""
    def __init__(self, service_id):
        super(DSManagerService, self).__init__(service_id)
        self.synchronizer = None
        self.add_rpc_endpoint(DSManagerEndpoints(self))

    def register_synchronizer(self):
        self.synchronizer = datasource_synchronizer.DatasourceSynchronizer(
            self.node)
        self.synchronizer.start()

    def start(self):
        super(DSManagerService, self).start()
        self.register_synchronizer()

    def stop(self):
        if self.synchronizer:
            self.synchronizer.stop()
        super(DSManagerService, self).stop()

    # Note(thread-safety): blocking function
    def add_datasource(self, item, deleted=False, update_db=True):
        req = self.make_datasource_dict(item)

        # check the request has valid information
        self.node.validate_create_datasource(req)
        if (len(req['name']) == 0 or req['name'][0] == '_'):
            raise exception.InvalidDatasourceName(value=req['name'])

        new_id = req['id']
        LOG.debug("adding datasource %s", req['name'])
        if update_db:
            LOG.debug("updating db")
            try:
                driver_info = self.node.get_driver_info(req['driver'])
                # Note(thread-safety): blocking call
                datasource = datasources_db.add_datasource(
                    id_=req['id'],
                    name=req['name'],
                    driver=req['driver'],
                    config=req['config'],
                    description=req['description'],
                    enabled=req['enabled'],
                    secret_config_fields=driver_info.get('secret', []))
            except db_exc.DBDuplicateEntry:
                raise exception.DatasourceNameInUse(value=req['name'])
            except db_exc.DBError:
                LOG.exception('Creating a new datasource failed due to '
                              'database backend error.')
                raise exception.DatasourceCreationError(value=req['name'])

        new_id = datasource['id']
        try:
            self.synchronizer.sync_datasource(req['name'])
            # immediate synch policies on local PE if present
            # otherwise wait for regularly scheduled synch
            engine = self.node.service_object(api_base.ENGINE_SERVICE_ID)
            if engine is not None:
                engine.synchronizer.sync_one_policy(req['name'])
            # TODO(dse2): also broadcast to all PE nodes to synch
        except exception.DataServiceError:
            LOG.debug('the datasource service is already '
                      'created in the node')
        except Exception:
            LOG.exception(
                'Unexpected exception encountered while registering '
                'new datasource %s.', req['name'])
            if update_db:
                datasources_db.delete_datasource(req['id'])
            msg = ("Datasource service: %s creation fails." % req['name'])
            raise exception.DatasourceCreationError(message=msg)

        new_item = dict(item)
        new_item['id'] = new_id
        return self.node.make_datasource_dict(new_item)

    # Note(thread-safety): blocking function
    def delete_datasource(self, datasource, update_db=True):
        LOG.debug("Deleting %s datasource ", datasource['name'])
        datasource_id = datasource['id']
        if update_db:
            # Note(thread-safety): blocking call
            result = datasources_db.delete_datasource_with_data(datasource_id)
            if not result:
                raise exception.DatasourceNotFound(id=datasource_id)

        # Note(thread-safety): blocking call
        try:
            self.synchronizer.sync_datasource(datasource['name'])
            # If local PE exists.. sync
            engine = self.node.service_object(api_base.ENGINE_SERVICE_ID)
            if engine:
                engine.synchronizer.sync_one_policy(datasource['name'])
        except Exception:
            msg = ('failed to synchronize_datasource after '
                   'deleting datasource: %s' % datasource_id)
            LOG.exception(msg)
            raise exception.DataServiceError(msg)


class DSManagerEndpoints(object):

    def __init__(self, service):
        self.ds_manager = service

    def add_datasource(self, context, items):
        return self.ds_manager.add_datasource(items)

    def delete_datasource(self, context, datasource):
        return self.ds_manager.delete_datasource(datasource)



# -*- coding: utf-8
#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Test for utils"""

import re

from oslo_log import log as logging

from congress.cfg_validator import utils
from congress.tests import base

LOG = logging.getLogger(__name__)


class TestUtils(base.TestCase):
    """Test of generic utility functions"""

    def test_hash(self):
        """Test shape of hash generated"""
        re_hash = ('^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-'
                   '[0-9a-f]{4}-[0-9a-f]{8}')
        self.assertTrue(re.match(re_hash, utils.compute_hash('foo')))
        self.assertTrue(re.match(re_hash, utils.compute_hash('foo', 'bar')))

    def test_cfg_value_to_congress(self):
        """Test sanitization of values for congress"""
        self.assertEqual('aAdef', utils.cfg_value_to_congress('aA%def'))
        self.assertEqual(u'aAf', utils.cfg_value_to_congress(u'aAéf'))
        # Do not replace 0 with ''
        self.assertEqual(0, utils.cfg_value_to_congress(0))
        # Do not replace 0.0 with ''
        self.assertEqual(0.0, utils.cfg_value_to_congress(0.0))
        self.assertEqual('', utils.cfg_value_to_congress(None))

    def test_add_rules(self):
        """Test adding a rule via the bus"""
        class _bus(object):

            def rpc(self, svc, command, arg):
                "fake rpc"
                return {"svc": svc, "command": command, "arg": arg}

        res = utils.add_rule(_bus(), 'policy', ['r1', 'r2'])
        expected = {
            'arg': {'policy_rules_obj': {
                'kind': 'nonrecursive',
                'name': 'policy',
                'rules': ['r1', 'r2']}},
            'command': 'persistent_create_policy_with_rules',
            'svc': '__engine'}
        self.assertEqual(expected, res)



# Copyright (c) 2016 NTT All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import datetime
from oslo_log import log as logging

from congress.datasources import constants
from congress.datasources import datasource_driver

LOG = logging.getLogger(__name__)


class PushDriver(datasource_driver.PushedDataSourceDriver):
    """A DataSource Driver for pushing tuples of data.

    To use this driver, run the following API:

    PUT /v1/data-sources/<the driver id>/tables/<table id>/rows

    Still a work in progress, but intent is to allow a request body
    to be any list of lists where the internal lists all have
    the same number of elements.

    request body:
    [ [1,2], [3,4] ]
    """

    def __init__(self, name='', args=None):
        super(PushDriver, self).__init__(name, args=args)
        self._table_deps['data'] = ['data']

    @classmethod
    def get_schema(cls):
        schema = {}
        # Hardcode the tables for now.  Later, create the tables on the fly.
        # May be as easy as deleting the following line.
        schema['data'] = []
        return schema

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'push'
        result['description'] = ('Datasource driver that allows external '
                                 'systems to push data.')
        # TODO(masa): Remove the REQUIRED config once python-congressclient
        # has been able to retrieve non-dict object in config fields at
        # $ openstack congress datasource list command
        result['config'] = {'description': constants.REQUIRED,
                            'persist_data': constants.OPTIONAL}
        return result

    def replace_entire_table_data(self, table_id, objs):
        LOG.info('update %s table in %s datasource', table_id, self.name)
        tablename = 'data'  # hard code
        self.prior_state = dict(self.state)
        self._update_state(tablename,
                           [tuple([table_id, tuple(x)]) for x in objs])
        LOG.debug('publish a new state %s in %s',
                  self.state[tablename], tablename)
        self.publish(tablename, self.state[tablename])
        self.number_of_updates += 1
        self.last_updated_time = datetime.datetime.now()



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import


from oslo_log import log as logging
from oslo_serialization import jsonutils as json

from congress.api import api_utils
from congress.api import base
from congress.api import error_codes
from congress.api import webservice
from congress import exception

LOG = logging.getLogger(__name__)


class DatasourceModel(base.APIModel):
    """Model for handling API requests about Datasources."""

    # Note(thread-safety): blocking function
    def get_items(self, params, context=None):
        """Get items in model.

        :param: params: A dict-like object containing parameters
            from the request query string and body.
        :param: context: Key-values providing frame of reference of request

        :returns: A dict containing at least a 'results' key whose value is
                 a list of items in the model.  Additional keys set in the
                 dict will also be rendered for the user.
        """

        # Note(thread-safety): blocking call
        results = self.bus.get_datasources(filter_secret=True)

        # Check that running datasources match the datasources in the
        # database since this is going to tell the client about those
        # datasources, and the running datasources should match the
        # datasources we show the client.

        return {"results": results}

    def get_item(self, id_, params, context=None):
        """Get datasource corresponding to id\_ in model."""
        try:
            datasource = self.bus.get_datasource(id_)
            return datasource
        except exception.DatasourceNotFound as e:
            LOG.debug("Datasource '%s' not found", id_)
            raise webservice.DataModelException(e.code, str(e),
                                                http_status_code=e.code)

    # Note(thread-safety): blocking function
    def add_item(self, item, params, id_=None, context=None):
        """Add item to model.

         :param: item: The item to add to the model
         :param: id\_: The ID of the item, or None if an ID should be generated
         :param: context: Key-values providing frame of reference of request

         :returns:  Tuple of (ID, newly_created_item)

         :raises  KeyError: ID already exists.
         """
        obj = None
        try:
            # Note(thread-safety): blocking call
            obj = self.invoke_rpc(base.DS_MANAGER_SERVICE_ID,
                                  'add_datasource',
                                  {'items': item},
                                  timeout=self.dse_long_timeout)
            # Let PE synchronizer take care of creating the policy.
        except (exception.BadConfig,
                exception.DatasourceNameInUse,
                exception.DriverNotFound,
                exception.DatasourceCreationError) as e:
            LOG.debug(_("Datasource creation failed."))
            raise webservice.DataModelException(
                e.code, webservice.original_msg(e), http_status_code=e.code)

        return (obj['id'], obj)

    # Note(thread-safety): blocking function
    def delete_item(self, id_, params, context=None):
        ds_id = context.get('ds_id')
        try:
            # Note(thread-safety): blocking call
            datasource = self.bus.get_datasource(ds_id)
            # FIXME(thread-safety):
            #  by the time greenthread resumes, the
            #  returned datasource name could refer to a totally different
            #  datasource, causing the rest of this code to unintentionally
            #  delete a different datasource
            #  Fix: check UUID of datasource before operating.
            #  Abort if mismatch
            self.invoke_rpc(base.DS_MANAGER_SERVICE_ID,
                            'delete_datasource',
                            {'datasource': datasource},
                            timeout=self.dse_long_timeout)
            # Let PE synchronizer takes care of deleting policy
        except (exception.DatasourceNotFound,
                exception.DanglingReference) as e:
            raise webservice.DataModelException(e.code, str(e))

    # Note(thread-safety): blocking function
    def request_refresh_action(self, params, context=None, request=None):
        caller, source_id = api_utils.get_id_from_context(context)
        try:
            args = {'source_id': source_id}
            # Note(thread-safety): blocking call
            self.invoke_rpc(caller, 'request_refresh', args)
        except exception.CongressException as e:
            LOG.debug(e)
            raise webservice.DataModelException.create(e)

    # Note(thread-safety): blocking function
    def execute_action(self, params, context=None, request=None):
        "Execute the action."
        service = context.get('ds_id')
        body = json.loads(request.body)
        action = body.get('name')
        action_args = body.get('args', {})
        if (not isinstance(action_args, dict)):
            (num, desc) = error_codes.get('execute_action_args_syntax')
            raise webservice.DataModelException(num, desc)

        try:
            args = {'service_name': service, 'action': action,
                    'action_args': action_args}
            # TODO(ekcs): perhaps keep execution synchronous when explicitly
            #   called via API
            # Note(thread-safety): blocking call
            self.invoke_rpc(base.ENGINE_SERVICE_ID, 'execute_action', args)
        except exception.PolicyException as e:
            (num, desc) = error_codes.get('execute_error')
            raise webservice.DataModelException(num, desc + "::" + str(e))

        return {}



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Unit test for the marshalling of options"""
import mock

from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging

from congress.cfg_validator.agent import generator
from congress.tests import base

LOG = logging.getLogger(__name__)


class TestGenerator(base.TestCase):
    """Unit test for the marshalling of options"""

    @mock.patch("oslo_config.generator._list_opts")
    def test_encode(self, mock_list_opts):
        """Test the json encoding of options"""

        opt1 = cfg.StrOpt("o1"),
        opt2 = cfg.IntOpt("o2", default=10),
        opt3 = cfg.StrOpt("o3", default="a"),
        mock_list_opts.return_value = [
            ("ns", [("g1", [opt1, opt2]), ("g2", [opt3])])]
        namespace = "ns"

        def _mko(nam, typ, kind, defv):
            return (
                '{"advanced": false, "default": %s, "deprecated_for_removal":'
                ' false, "deprecated_opts": [], "deprecated_reason": null,'
                ' "deprecated_since": null, "dest": "%s", "help": null,'
                ' "kind": "%s", "metavar": null, "mutable": false,'
                ' "name": "%s", "positional": false, "required": false,'
                ' "sample_default": null, "secret": false, "short": null,'
                ' "type": {"type": "%s"}}') % (defv, nam, kind, nam, typ)
        ns_string = (
            ('{"DEFAULT": {"namespaces": [], "object": null}, "g1": '
             '{"namespaces": [["ns", [[') +
            _mko("o1", "String", "StrOpt", "null") +
            '], [' + _mko("o2", "Integer", "IntOpt", 10) +
            ']]]], "object": null}, "g2": {"namespaces": [["ns", [[' +
            _mko("o3", "String", "StrOpt", "\"a\"") + ']]]], "object": null}}')
        computed = generator.generate_ns_data(namespace)
        self.assertEqual(
            ns_string, computed,
            "not the expected encoding of namespace")

    @mock.patch("oslo_config.generator._list_opts")
    def _test_encode_specific(self, opt, expected, mock_list_opts=None):
        """Test an option and check an expected string in result"""
        mock_list_opts.return_value = [
            ("ns", [("g", [opt])])]
        namespace = "ns"
        computed = generator.generate_ns_data(namespace)
        self.assertIn(expected, computed)

    def test_encode_range(self):
        """Test the json encoding of range option"""
        opt = cfg.Opt('o', types.Range(min=123, max=456))
        expected = '"type": {"max": 456, "min": 123, "type": "Range"}'
        self._test_encode_specific(opt, expected)

    def test_encode_list(self):
        """Test the json encoding of list option"""
        opt = cfg.ListOpt('opt', types.Boolean)
        expected = '"type": {"item_type": {"type": "String"}, "type": "List"}'
        self._test_encode_specific(opt, expected)

    def test_encode_dict(self):
        """Test the json encoding of dict option"""
        opt = cfg.DictOpt('opt')
        expected = '"type": {"type": "Dict", "value_type": {"type": "String"}}'
        self._test_encode_specific(opt, expected)

    def test_encode_ip(self):
        """Test the json encoding of IP option"""
        opt = cfg.IPOpt('opt')
        expected = '"type": {"type": "IPAddress"}'
        self._test_encode_specific(opt, expected)

    def test_encode_uri(self):
        """Test the json encoding of a URI"""
        opt = cfg.URIOpt('opt', 100)
        expected = '"type": {"max_length": 100, "type": "URI"}'
        self._test_encode_specific(opt, expected)

    def test_encode_regex(self):
        """Test the json encoding of a string option with regex"""
        opt = cfg.StrOpt('opt', regex=r'abcd')
        expected = '"type": {"regex": "abcd", "type": "String"}'
        self._test_encode_specific(opt, expected)

    def test_encode_deprecated(self):
        """Test Deprecated Opt: this is not the way to use it"""
        dep = cfg.DeprecatedOpt('o_old', 'g_old')
        opt = cfg.Opt('opt', deprecated_opts=[dep])
        expected = '"deprecated_opts": [{"group": "g_old", "name": "o_old"}]'
        self._test_encode_specific(opt, expected)



# Copyright (c) 2013 VMware, Styra.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock
import tenacity
import time

# Note(ekcs): this is needed for direct unit test because Dse2Runtime import,
#             which takes place before the confFixture is setup, fails w/o it
from novaclient import client as nova_client
from oslo_config import cfg
cfg.CONF.datasource_sync_period = 0
from oslo_messaging import conffixture

from congress.api import base as api_base
from congress.datalog import base as datalog_base
from congress.datalog import compile
from congress.datasources import nova_driver
from congress import exception as congressException
from congress.policy_engines import agnostic
from congress.tests.api import base as test_api_base
from congress.tests import base
from congress.tests import fake_datasource
from congress.tests import helper


class TestDSE(base.TestCase):

    def setUp(self):
        super(TestDSE, self).setUp()
        mc_fixture = conffixture.ConfFixture(cfg.CONF)
        mc_fixture.conf.transport_url = 'kombu+memory://'
        self.messaging_config = mc_fixture.conf
        self.messaging_config.rpc_response_timeout = 1

    def test_intranode_pubsub(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = fake_datasource.FakeDataSource('test1')
        test2 = fake_datasource.FakeDataSource('test2')
        node.register_service(test1)
        node.register_service(test2)

        test1.subscribe('test2', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test2.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))
        node.stop()

    def test_intranode_pubsub2(self):
        # same as test_intranode_pubsub but with opposite ordering.
        # (Ordering does matter with internode_pubsub).
        node = helper.make_dsenode_new_partition('testnode')
        test1 = fake_datasource.FakeDataSource('test1')
        test2 = fake_datasource.FakeDataSource('test2')
        node.register_service(test1)
        node.register_service(test2)

        test2.subscribe('test1', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test2, 'last_msg'), True)
        test1.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: test2.last_msg['data'], 42)
        self.assertFalse(hasattr(test1, "last_msg"))
        node.stop()

    def test_intranode_partial_unsub(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = fake_datasource.FakeDataSource('test1')
        test2 = fake_datasource.FakeDataSource('test2')
        node.register_service(test1)
        node.register_service(test2)

        test1.subscribe('test2', 'p')
        test1.subscribe('test2', 'q')
        test1.unsubscribe('test2', 'q')  # unsub from q should not affect p
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test2.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))
        node.stop()

    def test_sub_before_service_exists(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = fake_datasource.FakeDataSource('test1')
        node.register_service(test1)

        test1.subscribe('test2', 'p')
        self.assertFalse(hasattr(test1, "last_msg"))
        test2 = fake_datasource.FakeDataSource('test2')
        node.register_service(test2)
        test2.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))
        node.stop()
        node.wait()

    def test_internode_pubsub(self):
        node1 = helper.make_dsenode_new_partition('testnode1')
        test1 = fake_datasource.FakeDataSource('test1')
        node1.register_service(test1)
        node2 = helper.make_dsenode_same_partition(node1, 'testnode2')
        test2 = fake_datasource.FakeDataSource('test2')
        node2.register_service(test2)

        test1.subscribe('test2', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test2.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))
        node1.stop()
        node2.stop()

    def test_internode_partial_unsub(self):
        node1 = helper.make_dsenode_new_partition('testnode1')
        node2 = helper.make_dsenode_same_partition(node1, 'testnode2')
        test1 = fake_datasource.FakeDataSource('test1')
        test2 = fake_datasource.FakeDataSource('test2')
        node1.register_service(test1)
        node2.register_service(test2)

        test1.subscribe('test2', 'p')
        test1.subscribe('test2', 'q')
        test1.unsubscribe('test2', 'q')  # unsub from q should not affect p
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test2.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))
        node1.stop()
        node2.stop()

    def test_multiservice_pubsub(self):
        node1 = helper.make_dsenode_new_partition('testnode1')
        test1 = fake_datasource.FakeDataSource('test1')
        test2 = fake_datasource.FakeDataSource('test2')
        node1.register_service(test1)
        node1.register_service(test2)
        node2 = helper.make_dsenode_same_partition(node1, 'testnode2')
        test3 = fake_datasource.FakeDataSource('test3')
        node2.register_service(test3)

        test1.subscribe('test3', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test3.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))
        self.assertFalse(hasattr(test3, "last_msg"))
        node1.stop()
        node2.stop()

    def test_subscribe_snapshot(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = fake_datasource.FakeDataSource('test1')
        test2 = fake_datasource.FakeDataSource('test2')
        node.register_service(test1)
        node.register_service(test2)

        test1.subscribe('test2', 'fake_table')
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        self.assertEqual(test1.last_msg['data'], test2.state['fake_table'])
        node.stop()

    @mock.patch.object(nova_client, 'Client', spec_set=True, autospec=True)
    def test_datasource_sub(self, nova_mock):
        node = helper.make_dsenode_new_partition('testnode')
        nova = nova_driver.NovaDriver(
            name='nova', args=helper.datasource_openstack_args())
        test = fake_datasource.FakeDataSource('test')
        node.register_service(nova)
        node.register_service(test)

        nova.subscribe('test', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(nova, 'last_msg'), True)
        test.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: nova.last_msg['data'], 42)
        self.assertFalse(hasattr(test, "last_msg"))
        node.stop()

    @mock.patch.object(nova_client, 'Client', spec_set=True, autospec=True)
    def test_datasource_unsub(self, nova_mock):
        node = helper.make_dsenode_new_partition('testnode')
        nova = nova_driver.NovaDriver(
            name='nova', args=helper.datasource_openstack_args())
        test = fake_datasource.FakeDataSource('test')
        node.register_service(nova)
        node.register_service(test)

        nova.subscribe('test', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(nova, 'last_msg'), True)
        test.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: nova.last_msg['data'], 42)
        self.assertFalse(hasattr(test, "last_msg"))
        nova.unsubscribe('test', 'p')
        test.publish('p', 43, use_snapshot=True)
        # hard to test that the message is never delivered
        time.sleep(0.2)
        self.assertEqual(nova.last_msg['data'], 42)
        node.stop()

    @mock.patch.object(nova_client, 'Client', spec_set=True, autospec=True)
    def test_datasource_pub(self, nova_mock):
        node = helper.make_dsenode_new_partition('testnode')
        nova = nova_driver.NovaDriver(
            name='nova', args=helper.datasource_openstack_args())
        test = fake_datasource.FakeDataSource('test')
        node.register_service(nova)
        node.register_service(test)

        test.subscribe('nova', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test, 'last_msg'), True)
        nova.publish('p', 42, use_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: test.last_msg['data'], 42)
        self.assertFalse(hasattr(nova, "last_msg"))
        node.stop()

    def test_auto_resub(self):
        config = test_api_base.setup_config(with_fake_datasource=False,
                                            api=False, policy=False)
        node = config['node']
        config['ds_manager'].synchronizer.start()
        sub = fake_datasource.FakeDataSource('sub')
        pub = fake_datasource.FakeDataSource('pub')
        node.register_service(sub)
        node.register_service(pub)
        sub.subscribe('pub', 'p')

        helper.retry_check_function_return_value(
            lambda: hasattr(sub, 'last_msg'), True)
        helper.retry_check_function_return_value(
            lambda: sub.last_msg['data'], set([]))

        sub.receive_data_sequenced(
            'pub', 'p', [[1, 1]], 1, is_snapshot=True)
        helper.retry_check_function_return_value(
            lambda: sub.last_msg['data'], set([(1, 1)]))
        # skipping seqnum 2
        sub.receive_data_sequenced(
            'pub', 'p', [[3, 3]], 3, is_snapshot=True)
        # check that out-of-sequence update not applied
        self.assertRaises(
            tenacity.RetryError,
            helper.retry_check_function_return_value,
            lambda: sub.last_msg['data'], set([(3, 3)]))
        # check that resub takes place, setting data to initial state
        helper.retry_check_function_return_value(
            lambda: sub.last_msg['data'], set([]))
        node.stop()

    def test_datasource_poll(self):
        node = helper.make_dsenode_new_partition('testnode')
        pub = fake_datasource.FakeDataSource('pub')
        sub = fake_datasource.FakeDataSource('sub')
        node.register_service(pub)
        node.register_service(sub)

        sub.subscribe('pub', 'fake_table')
        pub.state = {'fake_table': set([(1, 2)])}
        pub.poll()

        helper.retry_check_function_return_value(
            lambda: sub.last_msg,
            {'publisher': 'pub',
             'data': (set(pub.state['fake_table']), set([])),
             'table': 'fake_table'})
        self.assertFalse(hasattr(pub, "last_msg"))
        node.stop()

    def test_policy_data(self):
        """Test policy correctly processes initial data snapshot."""
        node = helper.make_dsenode_new_partition('testnode')
        data = fake_datasource.FakeDataSource('data')
        engine = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        node.register_service(data)
        node.register_service(engine)

        engine.create_policy('policy1')
        engine.create_policy('data', kind=datalog_base.DATASOURCE_POLICY_TYPE)
        self.insert_rule(engine, 'p(x) :- data:fake_table(x)', 'policy1')
        data.state = {'fake_table': set([(1,), (2,)])}
        data.poll()
        helper.retry_check_db_equal(
            engine, 'p(x)', 'p(1) p(2)', target='policy1')
        self.assertFalse(hasattr(engine, "last_msg"))
        node.stop()

    def test_policy_data_update(self):
        """Test policy correctly processes initial data snapshot and update."""
        node = helper.make_dsenode_new_partition('testnode')
        data = fake_datasource.FakeDataSource('data')
        engine = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        node.register_service(data)
        node.register_service(engine)

        engine.create_policy('policy1')
        engine.create_policy('data', kind=datalog_base.DATASOURCE_POLICY_TYPE)
        self.insert_rule(engine, 'p(x) :- data:fake_table(x)', 'policy1')
        data.state = {'fake_table': set([(1,), (2,)])}
        data.poll()
        helper.retry_check_db_equal(
            engine, 'p(x)', 'p(1) p(2)', target='policy1')
        data.state = {'fake_table': set([(1,), (2,), (3,)])}
        data.poll()
        helper.retry_check_db_equal(
            engine, 'p(x)', 'p(1) p(2) p(3)', target='policy1')
        self.assertFalse(hasattr(engine, "last_msg"))
        node.stop()

    def test_policy_data_late_sub(self):
        """Test policy correctly processes data on late subscribe."""
        node = helper.make_dsenode_new_partition('testnode')
        data = fake_datasource.FakeDataSource('data')
        engine = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        node.register_service(data)
        node.register_service(engine)

        engine.create_policy('policy1')
        engine.create_policy('data', kind=datalog_base.DATASOURCE_POLICY_TYPE)
        data.state = {'fake_table': set([(1,), (2,)])}
        data.poll()
        self.insert_rule(engine, 'p(x) :- data:fake_table(x)', 'policy1')
        helper.retry_check_db_equal(
            engine, 'p(x)', 'p(1) p(2)', target='policy1')
        data.state = {'fake_table': set([(1,), (2,), (3,)])}
        data.poll()
        helper.retry_check_db_equal(
            engine, 'p(x)', 'p(1) p(2) p(3)', target='policy1')
        self.assertFalse(hasattr(engine, "last_msg"))
        node.stop()

    def insert_rule(self, engine, statement, target=None):
        statement = compile.parse1(statement)
        if target is None:
            e = compile.Event(statement)
        else:
            e = compile.Event(statement, target=target)
        engine.process_policy_update([e])

    def test_unregister(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = fake_datasource.FakeDataSource('test1')
        node.register_service(test1)
        obj = node.invoke_service_rpc(
            'test1', 'get_status', {'source_id': None, 'params': None})
        self.assertIsNotNone(obj)
        node.unregister_service('test1')
        helper.retry_til_exception(
            congressException.NotFound,
            lambda: node.invoke_service_rpc(
                'test1', 'get_status', {'source_id': None, 'params': None}))
        node.stop()

    def _create_node_with_services(self, nodes, services, num, partition_id):
        nid = 'cbd_node%s' % num
        nodes.append(helper.make_dsenode_same_partition(partition_id, nid))
        ns = []
        for s in range(num):
            # intentionally starting different number services
            ns.append(
                fake_datasource.FakeDataSource('cbd-%d_svc-%d' % (num, s)))
            nodes[-1].register_service(ns[-1])
        services.append(ns)
        return nodes[-1]

    def test_subs_list_update_aggregated_by_service(self):
        part = helper.get_new_partition()
        nodes = []
        services = []
        num_nodes = 3

        for i in range(num_nodes):
            n = self._create_node_with_services(nodes, services, i, part)
            n.start()

        # add subscriptions
        for i in range(2, num_nodes):
            for s2 in services[i]:
                for s1 in services[i-1]:
                    s1.subscribe(s2.service_id, 'table-A')
                    s2.subscribe(s1.service_id, 'table-B')
        services[1][0].subscribe(services[2][0].service_id, 'table-C')
        services[2][1].subscribe(services[2][0].service_id, 'table-D')

        # constructed expected results
        expected_subbed_tables = {}
        expected_subbed_tables[nodes[1].node_id] = {}
        expected_subbed_tables[nodes[2].node_id] = {}
        expected_subbed_tables[nodes[1].node_id][
            services[1][0].service_id] = set(['table-B'])
        expected_subbed_tables[nodes[2].node_id][
            services[2][0].service_id] = set(['table-A', 'table-C', 'table-D'])
        expected_subbed_tables[nodes[2].node_id][
            services[2][1].service_id] = set(['table-A'])

        # validate
        def _validate_subbed_tables(node):
            for s in node.get_services():
                sid = s.service_id
                subscribed_tables = node.service_object(
                    sid)._published_tables_with_subscriber
                self.assertEqual(
                    subscribed_tables,
                    expected_subbed_tables[node.node_id][sid],
                    '%s has incorrect subscribed tables list' % sid)
            return True
        for n in nodes:
            helper.retry_check_function_return_value(
                lambda: _validate_subbed_tables(n), True)

        # selectively unsubscribe
        services[1][0].unsubscribe(services[2][0].service_id, 'table-A')
        # note that services[2][1] still subscribes to 'table-B'
        services[2][0].unsubscribe(services[1][0].service_id, 'table-B')
        # extraneous unsubscribe
        services[2][0].unsubscribe(services[1][0].service_id, 'table-None')

        # update expected results
        expected_subbed_tables[nodes[2].node_id][
            services[2][0].service_id] = set(['table-C', 'table-D'])

        for n in nodes:
            helper.retry_check_function_return_value(
                lambda: _validate_subbed_tables(n), True)

        # resubscribe
        services[1][0].subscribe(services[2][0].service_id, 'table-A')
        services[2][0].subscribe(services[1][0].service_id, 'table-B')

        # update expected results
        expected_subbed_tables[nodes[2].node_id][
            services[2][0].service_id] = set(['table-A', 'table-C', 'table-D'])

        for n in nodes:
            helper.retry_check_function_return_value(
                lambda: _validate_subbed_tables(n), True)

    # TODO(dse2): the policy table subscription feature is not ready in DSE2
    #     the problem is that compile and agnostic are not ready to deal with
    #     the double qualifier publishing_policy_engine:policy_name:table_name
    #     that the receiving policy engine needs to handle when subscribing to
    #     policy_name:table_name from publishing_policy_engine.
    #     This feature is not urgent because Congress currently does not have
    #     one policy engine subscribing from another policy engine
    #     (only from data source)
    # def test_policy_table_publish(self):
    #     """Policy table result publish
    #
    #     Test basic DSE functionality with policy engine and table result
    #     publish.
    #     """
    #     node = helper.make_dsenode_new_partition('testnode')
    #     data = fake_datasource.FakeDataSource('data')
    #     policy = agnostic.DseRuntime('policy')
    #     policy2 = agnostic.DseRuntime('policy2')
    #     node.register_service(data)
    #     node.register_service(policy)
    #     node.register_service(policy2)
    #     policy.synchronizer = mock.MagicMock()
    #     policy2.synchronizer = mock.MagicMock()
    #
    #     policy.create_policy(
    #         'data', kind=datalog_base.DATASOURCE_POLICY_TYPE)
    #     policy.create_policy('classification')
    #     policy.set_schema('data', compile.Schema({'q': (1,)}))
    #     policy.insert('p(x):-data:q(x),gt(x,2)', target='classification')
    #
    #     policy.insert('q(3)', target='data')
    #     # TODO(ekcs): test that no publish triggered (because no subscribers)
    #
    #     policy2.create_policy('policy')
    #     policy2.subscribe('policy', 'classification:p')
    #     helper.retry_check_function_return_value(
    #         lambda: 'classification:p' in
    #         policy._published_tables_with_subscriber, True)
    #     self.assertEqual(list(policy.policySubData.keys()),
    #                      [('p', 'classification', None)])
    #
    #     helper.retry_check_db_equal(
    #         policy2, 'policy:classification:p(x)',
    #         'policy:classification:p(3)')
    #
    #     policy.insert('q(4)', target='data')
    #     helper.retry_check_db_equal(
    #         policy2, 'policy:classification:p(x)',
    #         ('policy:classification:p(3)'
    #          ' policy:classification:p(4)'))
    #
    #     # test that no change to p means no publish triggered
    #     policy.insert('q(2)', target='data')
    #     # TODO(ekcs): test no publish triggered
    #
    #     policy.delete('q(4)', target='data')
    #     helper.retry_check_db_equal(
    #         policy2, 'policy:classification:p(x)',
    #         'policy:classification:p(3)')
    #
    #     policy2.unsubscribe('policy', 'classification:p')
    #     # trigger removed
    #     helper.retry_check_function_return_value(
    #         lambda: len(policy._published_tables_with_subscriber) == 0, True)
    #     self.assertEqual(list(policy.policySubData.keys()), [])
    #
    #     policy.insert('q(4)', target='data')
    #     # TODO(ekcs): test that no publish triggered (because no subscribers)
    #     node.stop()

    def test_replicated_pe_exec(self):
        """Test correct local leader behavior with 2 PEs requesting exec"""
        node1 = helper.make_dsenode_new_partition('testnode1')
        node2 = helper.make_dsenode_same_partition(node1, 'testnode2')
        dsd = fake_datasource.FakeDataSource('dsd')
        # faster time-out for testing
        dsd.LEADER_TIMEOUT = 2
        pe1 = agnostic.DseRuntime('pe1')
        pe2 = agnostic.DseRuntime('pe2')
        node1.register_service(pe1)
        node2.register_service(pe2)
        node1.register_service(dsd)
        assert dsd._running
        assert node1._running
        assert node2._running
        assert node1._control_bus._running

        # first exec request obeyed and leader set
        pe2.rpc('dsd', 'request_execute',
                {'action': 'fake_act', 'action_args': {'name': 'testnode2'},
                 'wait': True})
        helper.retry_check_function_return_value(
            lambda: len(dsd.exec_history), 1)
        self.assertEqual(dsd._leader_node_id, 'testnode2')

        # second exec request from leader obeyed and leader remains
        pe2.rpc('dsd', 'request_execute',
                {'action': 'fake_act', 'action_args': {'name': 'testnode2'},
                 'wait': True})
        helper.retry_check_function_return_value(
            lambda: len(dsd.exec_history), 2)
        self.assertEqual(dsd._leader_node_id, 'testnode2')

        # exec request from non-leader not obeyed
        pe1.rpc('dsd', 'request_execute',
                {'action': 'fake_act', 'action_args': {'name': 'testnode1'},
                 'wait': True})
        self.assertRaises(
            tenacity.RetryError,
            helper.retry_check_function_return_value,
            lambda: len(dsd.exec_history), 3)

        # leader vacated after heartbeat stops
        node2.stop()
        node2.wait()
        helper.retry_check_function_return_value(
            lambda: dsd._leader_node_id, None)

        # next exec request obeyed and new leader set
        pe1.rpc('dsd', 'request_execute',
                {'action': 'fake_act', 'action_args': {'name': 'testnode1'},
                 'wait': True})
        helper.retry_check_function_return_value(
            lambda: len(dsd.exec_history), 3)
        self.assertEqual(dsd._leader_node_id, 'testnode1')
        node1.stop()
        node2.stop()



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Utilities for testing RPC clients.
"""
import mock

from oslotest import base


# TODO(pc): all this does not ensure that transport and target make any sense.
# But not tested in neutron ml2 use case either.
class BaseTestRpcClient(base.BaseTestCase):
    """Abstract class providing functions to test client RPC"""

    def _test_rpc_api(self, rpcapi, topic, method, rpc_method, **kwargs):
        """Base function to test each call"""
        ctxt = {}
        expected_retval = 'foo' if rpc_method == 'call' else None
        expected_version = kwargs.pop('version', None)
        fanout = kwargs.pop('fanout', False)
        server = kwargs.get('server', None)

        with mock.patch.object(rpcapi.client, rpc_method) as rpc_mock,\
                mock.patch.object(rpcapi.client, 'prepare') as prepare_mock:
            prepare_mock.return_value = rpcapi.client
            rpc_mock.return_value = expected_retval
            retval = getattr(rpcapi, method)(ctxt, **kwargs)

        prepare_args = {}
        if expected_version:
            prepare_args['version'] = expected_version
        if fanout:
            prepare_args['fanout'] = fanout
        if topic:
            prepare_args['topic'] = topic
        if server:
            prepare_args['server'] = server
        prepare_mock.assert_called_once_with(**prepare_args)

        self.assertEqual(retval, expected_retval)
        if server:
            kwargs.pop('server')
        rpc_mock.assert_called_once_with(ctxt, method, **kwargs)



# Copyright (c) 2018 VMware Inc All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from datetime import datetime
from datetime import timedelta

import eventlet
from futurist import periodics
from oslo_concurrency import lockutils
from oslo_log import log as logging

from congress.datasources import constants
from congress.datasources import datasource_driver

LOG = logging.getLogger(__name__)

TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


class VitrageDriver(datasource_driver.PushedDataSourceDriver):
    '''Datasource driver that accepts Vitrage webhook alarm notification.'''

    value_trans = {'translation-type': 'VALUE'}

    def flatten_and_timestamp_alarm_webhook(webhook_alarms_objects):
        flattened = []
        key_to_sub_dict = 'resource'
        for alarm in webhook_alarms_objects:
            sub_dict = alarm.pop(key_to_sub_dict)
            for k, v in sub_dict.items():
                # add prefix to key and move to top level dict
                alarm[key_to_sub_dict + '_' + k] = v
            alarm['receive_timestamp'] = datetime.utcnow().strftime(
                TIMESTAMP_FORMAT)
            flattened.append(alarm)
        return flattened

    webhook_alarm_translator = {
        'translation-type': 'HDICT',
        'table-name': 'alarms',
        'selector-type': 'DICT_SELECTOR',
        'objects-extract-fn': flatten_and_timestamp_alarm_webhook,
        'field-translators':
            ({'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'vitrage_type', 'col': 'type',
              'translator': value_trans},
             {'fieldname': 'vitrage_operational_severity',
              'col': 'operational_severity',
              'translator': value_trans},
             {'fieldname': 'vitrage_id', 'translator': value_trans},
             {'fieldname': 'update_timestamp', 'translator': value_trans},
             {'fieldname': 'receive_timestamp', 'translator': value_trans},
             {'fieldname': 'resource_name', 'translator': value_trans},
             {'fieldname': 'resource_id', 'translator': value_trans},
             {'fieldname': 'resource_vitrage_id', 'translator': value_trans},
             {'fieldname': 'resource_project_id', 'translator': value_trans},
             {'fieldname': 'resource_vitrage_operational_state',
              'col': 'resource_operational_state',
              'translator': value_trans},
             {'fieldname': 'resource_vitrage_type',
              'col': 'resource_type', 'translator': value_trans},
             )}

    TRANSLATORS = [webhook_alarm_translator]

    def __init__(self, name='', args=None):
        LOG.warning(
            'The Vitrage driver is classified as having unstable schema. '
            'The schema may change in future releases in '
            'backwards-incompatible ways.')
        super(VitrageDriver, self).__init__(name, args=args)
        if args is None:
            args = {}
        # set default time to 10 days before deleting an active alarm
        self.hours_to_keep_alarm = int(args.get('hours_to_keep_alarm', 240))
        self.set_up_periodic_tasks()

    @lockutils.synchronized('congress_vitrage_ds_data')
    def _webhook_handler(self, payload):
        tablename = 'alarms'

        # remove previous alarms of same ID from table
        row_id = payload['payload']['vitrage_id']
        column_index_number_of_row_id = 4
        to_remove = [row for row in self.state[tablename]
                     if row[column_index_number_of_row_id] == row_id]
        for row in to_remove:
            self.state[tablename].discard(row)

        # add new alarm to table
        translator = self.webhook_alarm_translator
        row_data = VitrageDriver.convert_objs(
            [payload['payload']], translator)
        for table, row in row_data:
            if table == tablename:
                self.state[tablename].add(row)

        LOG.debug('publish a new state %s in %s',
                  self.state[tablename], tablename)
        # Note (thread-safety): blocking call
        self.publish(tablename, self.state[tablename])
        return [tablename]

    def set_up_periodic_tasks(self):
        @lockutils.synchronized('congress_vitrage_ds_data')
        @periodics.periodic(spacing=max(self.hours_to_keep_alarm * 3600/10, 1))
        def delete_old_alarms():
            tablename = 'alarms'
            col_index_of_timestamp = 5
            # find for removal all alarms at least self.hours_to_keep_alarm old
            to_remove = [
                row for row in self.state[tablename]
                if (datetime.utcnow() -
                    datetime.strptime(row[col_index_of_timestamp],
                                      TIMESTAMP_FORMAT)
                    >= timedelta(hours=self.hours_to_keep_alarm))]
            for row in to_remove:
                self.state[tablename].discard(row)

        periodic_task_callables = [
            (delete_old_alarms, None, {}),
            (delete_old_alarms, None, {})]
        self.periodic_tasks = periodics.PeriodicWorker(periodic_task_callables)
        self.periodic_tasks_thread = eventlet.spawn_n(
            self.periodic_tasks.start)

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'vitrage'
        result['description'] = ('Datasource driver that accepts Vitrage '
                                 'webhook alarm notifications.')
        result['config'] = {'persist_data': constants.OPTIONAL,
                            'hours_to_keep_alarm': constants.OPTIONAL}
        return result

    def __del__(self):
        if self.periodic_tasks:
            self.periodic_tasks.stop()
            self.periodic_tasks.wait()
            self.periodic_tasks = None
        if self.periodic_tasks_thread:
            eventlet.greenthread.kill(self.periodic_tasks_thread)
            self.periodic_tasks_thread = None
